'''
Deaggregation Management implementation

Stanley Bak
Nov 2018
'''

from collections import deque, namedtuple

from hylaa.util import Freezable
from hylaa.timerutil import Timers

# Operation types
OpInvIntersect = namedtuple('OpInvIntersect', ['step', 'node', 'i_index', 'is_stronger'])
OpLeftInvariant = namedtuple('OpLeftInvariant', ['step', 'node', 'reached_time_bound'])

class OpTransition(): # pylint: disable=too-few-public-methods
    'a transition operation'

    def __init__(self, step, parent_node, child_node, transition, poststate):
        self.step = step
        self.parent_node = parent_node
        self.child_node = child_node
        self.transition = transition
        self.poststate = poststate

    def __str__(self):
        return f"[OpTransition({self.step}, {self.transition})]"

class DeaggregationManager(Freezable):
    'manager for deaggregation data'

    def __init__(self, aggdag):
        self.aggdag = aggdag
        
        self.waiting_nodes = deque() # a deque of 2-tuples (parent, children_list) of AggDagNodes awaiting deaggregation

        # associated with the current computaiton
        self.deagg_parent = None # a AggDagNode
        self.deagg_children = None # a list of AggDagNode

        # during replay, transition sucessors may have common children nodes that got aggregated
        # this maps the overapproximation node to the new unaggregated components (list of stateset and list of ops)
        # str(AggDagNode.id()) -> list_of_TransitionOps 
        self.nodes_to_ops = None # maps (newop.parent, oldop.child) to list of transtion_ops

        self.replay_step = None # current step during a replay

    def doing_replay(self):
        'are we in the middle of a refinement replay?'

        return self.replay_step is not None or self.waiting_nodes

    def plot_replay_deaggregated(self, cur_step_in_mode):
        '''plot the updated states during a replay

        child_plot_mask is a list of booleans, indicating which deagg chidlren should be plotted.
        plots may be skipped if the child left the invariant on an earlier step

        note: invariant can be false for one step
        '''

        plotman = self.aggdag.core.plotman

        # add children plots
        plot_child_states = []
        
        for i, child in enumerate(self.deagg_children):
            if not child.node_left_invariant():
                plot_child_states.append(child.stateset)
                assert child.stateset.cur_step_in_mode == cur_step_in_mode, \
                  f"expected cur_step_in_mode {cur_step_in_mode}, but it was {child.stateset.cur_step_in_mode}"

        plotman.add_plotted_states(plot_child_states)

        # delete parent plot
        plotman.delete_plotted_state(self.deagg_parent.stateset, cur_step_in_mode)

    def init_replay(self):
        'initialize a replay action'

        assert self.replay_step is None

        self.deagg_parent, self.deagg_children = self.waiting_nodes.popleft()
        self.replay_step = 0
        self.nodes_to_ops = {}

        # pause plot
        self.aggdag.core.print_verbose("Pausing due to start of do_step_replay()")
        self.aggdag.core.plotman.pause()

    def do_step_replay(self):
        'do a refinement replay step'

        assert self.doing_replay()

        if self.replay_step is None:
            self.init_replay()
        
        assert self.replay_step is not None

        op_list = self.deagg_parent.op_list
        op = op_list[self.replay_step]
        cur_step_in_mode = op.step

        # This part is a bit complicated due to discrete-time sematnics.
        # Basically, in continuous post the operation order is:
        # 1. inv intersect
        # 2. step()
        # 3. check transitions
        # 4. plot
        #
        # in order to recreate this during a replay, we need conditional branching:
        # if first op is transition, process all transitions [which does step() first] at the current step,
        #                                                    but no invariant intersections
        # if first op is invariant intersection, process all ops at the current step and transitions at the next step
        #  
        only_process_transitions = isinstance(op_list[self.replay_step], OpTransition)

        parent_left_invariant = False
        
        for op in op_list[self.replay_step:]:
            if op.step > cur_step_in_mode and not only_process_transitions:
                cur_step_in_mode += 1
                only_process_transitions = True
            
            if op.step != cur_step_in_mode:
                #print(".deagg breaking because next op step is not cur_step_in_mode")
                break

            if only_process_transitions and not isinstance(op, OpTransition):
                #print(f".deagg breaking because next op is not OpTransition: {op}")
                break

            #print(f".deagg replaying step {self.replay_step}/{len(self.deagg_parent.op_list)}: {op}")

            if isinstance(op, OpLeftInvariant):
                parent_left_invariant = True

            for child in self.deagg_children:
                if not child.node_left_invariant():
                    child.replay_op(self.deagg_parent.op_list, self.replay_step)
                else:
                    self.aggdag.core.print_verbose("child has left invariant")

            self.replay_step += 1

        # advance the children to the current step if in case we only had inv_intersection operations
        for child in self.deagg_children:
            if not child.node_left_invariant():
                child.stateset.step(cur_step_in_mode)

        if not parent_left_invariant:
            # do the plot after the replay actions, as this can affect the plot (for example, invariant intersection)
            self.plot_replay_deaggregated(cur_step_in_mode)

        was_last_step = self.replay_step >= len(self.deagg_parent.op_list)

        if was_last_step or all([c.node_left_invariant() for c in self.deagg_children]):
            # update recursive children
            
            for pair, ops in self.nodes_to_ops.items():
                _, child = pair
                self.aggdag.core.print_verbose(f"making node for recursive deaggregation with t={ops[0].transition}")

                # aggregate all ops into a single node, using same aggregation as before
                node = self.aggdag.make_node(ops, child.agg_type_from_parents, child.stateset.aggstring)
                
                self.waiting_nodes.append((child, [node]))
                
            self.deagg_parent = self.deagg_children = self.replay_step = self.nodes_to_ops = None

            #print(".deaggregation calling aggdag.save_viz()")
            #self.aggdag.save_viz()

    def update_transition_successors(self, old_op, new_op):
        '''
        during a replay, update a node's successsors recursively
        '''

        # aggregation will only be done if both new_op.parent and old_op.child match
        parent_node = new_op.parent_node
        child_node = old_op.child_node
        pair = (parent_node, child_node)

        if pair not in self.nodes_to_ops:
            ops = []
            self.nodes_to_ops[pair] = ops
        else:
            ops = self.nodes_to_ops[pair]

        ops.append(new_op)

    def begin_replay(self, node):
        'begin a deaggregation replay with the passed-in node'

        Timers.tic('begin deagg replay')

        # remove all states in the waiting list that come from this node
        self.aggdag.remove_node_decendants_from_waiting_list(node)

        # start to populate waiting_nodes
        self.waiting_nodes.append((node, node.split()))

        Timers.toc('begin deagg replay')



'''
Hybrid Automaton generic definition for Hylaa
Stanley Bak (Sept 2016)
'''

import numpy as np

import scipy as sp
from scipy.sparse import csr_matrix, csc_matrix

from hylaa.util import Freezable
from hylaa.time_elapse import TimeElapser

from hylaa import lputil

class LinearConstraint(Freezable):
    'a single csr sparse linear constraint: csr_vec * x <= rhs'

    def __init__(self, csr_vec, rhs):
        if not isinstance(csr_vec, csr_matrix):
            csr_vec = csr_matrix(csr_vec, dtype=float)
            
        assert csr_vec.shape[0] == 1, "expected single row constraint vector"
        
        self.csr = csr_vec
        self.rhs = float(rhs)

        self.freeze_attrs()

    def almost_equals(self, other, tol):
        'equality up to a tolerance'
        assert isinstance(other, LinearConstraint)

        rv = True

        if abs(self.rhs - other.rhs) > tol:
            rv = False
        elif self.csr.shape != other.csr.shape:
            rv = False
        elif not (self.csr.indices == other.csr.indices).all():
            rv = False
        elif not (self.csr.indptr == self.csr.indptr).all():
            rv = False
        else:
            for i in range(len(self.csr.data)):
                a = self.csr.data[i]
                b = other.csr.data[i]

                if abs(a - b) > tol:
                    rv = False
                    break

        return rv

    def is_true_for_point(self, point):
        'is the constraint true for the passed in point?'

        return self.csr * point <= self.rhs

    def negate(self):
        'return the negation of the condition'

        return LinearConstraint(-1 * self.csr, -self.rhs)

    def __str__(self):
        return '[LinearConstraint: {} * x <= {}]'.format(self.csr.toarray(), self.rhs)

    def __repr__(self):
        return 'LinearConstraint({}, {})'.format(repr(self.csr), repr(self.rhs))

class Mode(Freezable):
    '''
    A single mode of a hybrid automaton with dynamics x' = Ax + b. 

    The dynamics should be set_dynamics and (optionally) set_inputs. If they are not set, the mode will be
    considered an error mode.
    '''

    def __init__(self, ha, name, mode_id):
        assert isinstance(ha, HybridAutomaton)

        self.ha = ha # pylint: disable=invalid-name
        self.name = name
        self.mode_id = mode_id # unique int identified for this mode

        # dynamics are x' = Ax + Bu
        self.a_csr = None
        self.b_csr = None

        # constraints on input
        self.u_constraints_csc = None # csc_matrix
        self.u_constraints_rhs = None # 1-d np.ndarray

        self.transitions = [] # outgoing Transition objects

        self.inv_list = [] # a list of LinearConstraint, if all are true then the invariant is true

        self.time_elapse = None # a TimeElapse object... initialized on init_time_elapse()

        self.freeze_attrs()

    def is_error(self):
        'is this an error mode'

        return self.a_csr is None

    def point_in_invariant(self, point):
        'is the passed-in point inside the invariant?'

        all_true = True

        for lc in self.inv_list:
            if not lc.is_true_for_point(point):
                all_true = False
                break

        return all_true

    def set_invariant(self, constraints_csr, constraints_rhs):
        'sets the invariant'

        assert self.a_csr is not None, "A matrix must be set first"
        
        if not isinstance(constraints_csr, csr_matrix):
            constraints_csr = csr_matrix(constraints_csr, dtype=float)
            
        if not isinstance(constraints_rhs, np.ndarray):
            constraints_rhs = np.array(constraints_rhs, dtype=float)

        constraints_rhs.shape = (len(constraints_rhs), ) # flatten rhs into a 1-d array
        assert constraints_csr.shape[1] == self.a_csr.shape[0], \
            "width of invaraiant constraints({}) must equal A matrix size({})".format( \
            constraints_csr.shape[1], self.a_csr.shape[0])
        assert constraints_csr.shape[0] == len(constraints_rhs)

        assert lputil.is_feasible(constraints_csr, constraints_rhs), "invariant constraints were infeasible"

        # for efficiency in checking, the invariant is split into a list of individual constraints
        for row, rhs in enumerate(constraints_rhs):
            inds = []
            data = []
            indptr = [0]

            for i in range(constraints_csr.indptr[row], constraints_csr.indptr[row+1]):
                column = constraints_csr.indices[i]
                d = constraints_csr.data[i]

                inds.append(column)
                data.append(d)

            indptr.append(len(data))

            constraint_vec = csr_matrix((data, inds, indptr), dtype=float, shape=(1, constraints_csr.shape[1]))
            
            self.inv_list.append(LinearConstraint(constraint_vec, rhs))

    def _check_inputs(self, b_csr, u_constraints_csr, u_constraints_rhs, allow_constants):
        'Run assersion checks on input matrices'

        assert u_constraints_csr.shape[0] == u_constraints_rhs.shape[0], "u_constraints rows shoud match rhs len"
        assert u_constraints_csr.shape[1] == b_csr.shape[1], "u_constraints cols should match b.width"

        assert b_csr.shape[0] == self.a_csr.shape[0], \
                "B-mat shape {} incompatible with A-mat shape {}".format(b_csr.shape, self.a_csr.shape)

        # make sure the input constraints are feasible
        assert lputil.is_feasible(u_constraints_csr, u_constraints_rhs), "input constraints were infeasible"

        #make sure there are not inputs that are fixed to a constant. This is for efficiency reasons. It is better 
        #to add an affine variable to the a matrix and including this as part of A.
        
        num_inputs = b_csr.shape[1]

        for i in range(num_inputs):
            # does this input only affect a single variable? --> does b_col have a single nonzero?
            b_col = b_csr[:, i].toarray()
            nonzeros = sum([1 if x != 0 else 0 for x in b_col])

            if nonzeros != 1:
                continue

            # check if is there a fixed lower and upper bound for this input
            lb_row = np.array([0 if n != i else 1 for n in range(num_inputs)], dtype=float)
            ub_row = np.array([0 if n != i else -1 for n in range(num_inputs)], dtype=float)
            lb = ub = None

            for row, rhs in zip(u_constraints_csr, u_constraints_rhs):
                row = row.toarray()
                if np.array_equiv(lb_row, row):
                    lb = rhs
                elif np.array_equiv(ub_row, row):
                    ub = -rhs

            if ub is None or lb is None:
                continue

            if abs(ub-lb) < 1e-9:
                msg = ("Time-varying input #{} is fixed to {}. This is a (very) inefficient " + \
                    "way encode affine terms. Instead, introduce a fixed affine variable in the A matrix with a' = 0" + \
                    " and a(0) = 1, and refer to that variable in any differential equations that use affine " + \
                    "terms. This can be made a warning by using 'allow_constants=True' in set_inputs().").format(i, lb)   
                             
                if not allow_constants:
                    raise RuntimeError(msg)
                else:
                    print("Warning: " + msg) 

    def set_inputs(self, b_csr, u_constraints_csr, u_constraints_rhs, allow_constants=False):
        '''sets the time-varying / uncertain inputs for the mode (optional)

        if allow_constants is True, this will permit inputs that are fixed to constants. This is inefficient though,
        you should instead add an affine variables to the A matrix that's initially equal to 1 with derivative 0, and
        refer to that variable in the A matrix, rather than adding inputs.
        '''
    
        assert self.a_csr is not None, "set_dynamics should be done before set_inputs"
        if not isinstance(b_csr, csr_matrix):
            b_csr = csr_matrix(b_csr, dtype=float)
            
        if not isinstance(u_constraints_csr, csr_matrix):
            u_constraints_csr = csr_matrix(u_constraints_csr, dtype=float)
            
        if not isinstance(u_constraints_rhs, np.ndarray):
            u_constraints_rhs = np.array(u_constraints_rhs, dtype=float)
        
        u_constraints_rhs.shape = (len(u_constraints_rhs), ) # flatten init_rhs into a 1-d array

        # additional checks on inputs
        self._check_inputs(b_csr, u_constraints_csr, u_constraints_rhs, allow_constants)

        self.b_csr = b_csr

        # inputs will be assigned column-by-column, so store constraints in csc matrices
        self.u_constraints_csc = csc_matrix(u_constraints_csr)
        self.u_constraints_rhs = u_constraints_rhs

    def set_dynamics(self, a_csr):
        'sets the autonomous system dynamics (A matrix)'

        if not isinstance(a_csr, csr_matrix):
            a_csr = csr_matrix(a_csr, dtype=float)

        assert a_csr.shape[0] == a_csr.shape[1], "expected square dynamics matrix, got {}".format(a_csr.shape)

        self.a_csr = a_csr

    def init_time_elapse(self, step_size):
        'initialize the time elapse object for this mode (called by verification core)'

        if self.a_csr is not None:
            self.time_elapse = TimeElapser(self, step_size)

    def __str__(self):
        return '[AutomatonMode with name:{}, a_matrix:{}]'.format(self.name, \
            "None" if self.a_csr is None else self.a_csr.toarray())

    def __repr__(self):
        return str(self)

class Transition(Freezable):
    'A transition of a hybrid automaton'

    def __init__(self, ha, from_mode, to_mode, name=''):
        assert isinstance(ha, HybridAutomaton)
        self.ha = ha # pylint: disable=invalid-name
        self.from_mode = from_mode
        self.to_mode = to_mode

        self.guard_csr = None
        self.guard_rhs = None

        self.reset_csr = None
        self.reset_minkowski_csr = None
        self.reset_minkowski_constraints_csr = None
        self.reset_minkowski_constraints_rhs = None

        self.name = name

        self.time_triggered = False # assigned automatically if settings.optimize_tt_transitions == True

        self.transition_index = len(from_mode.transitions)
        from_mode.transitions.append(self)

        self.freeze_attrs()

    def is_guard_true_for_point(self, pt):
        '''is the guard condition true for the given point?'''

        all_true = True

        val = self.guard_csr * pt

        for lhs, rhs in zip(val, self.guard_rhs):
            if lhs > rhs:
                all_true = False
                break

        return all_true

    def apply_reset_for_point(self, pt):
        '''apply the reset for the passed-in point

        returns new_point, new_mode
        '''

        assert self.reset_minkowski_csr is None, "resets with minkowski terms are unimplemented"

        new_mode = self.to_mode

        if self.reset_csr is None:
            new_pt = pt
        else:
            new_pt = self.reset_csr * pt

        return new_pt, new_mode

    def set_guard_true(self):
        '''sets the guard to be True (always enabled)'''

        assert self.from_mode.a_csr is not None, "A-matrix not assigned in predecessor mode {}".format(self.from_mode)
        dims = self.from_mode.a_csr.shape[1]

        self.set_guard(csr_matrix((0, dims), dtype=float), [])

    def set_guard(self, guard_csr, guard_rhs):
        '''set the guard'''
        
        if not isinstance(guard_csr, csr_matrix):
            guard_csr = csr_matrix(guard_csr, dtype=float)
            
        if not isinstance(guard_rhs, np.ndarray):
            guard_rhs = np.array(guard_rhs, dtype=float)
        
        assert self.from_mode.a_csr is not None, "A-matrix not assigned in predecessor mode {}".format(self.from_mode)

        if guard_csr.shape[0] > 0:
            assert guard_csr.shape[1] == self.from_mode.a_csr.shape[0], "guard matrix expected {} columns, got {}" \
                                                          .format(self.from_mode.a_csr.shape[0], guard_csr.shape[1])

        assert lputil.is_feasible(guard_csr, guard_rhs), "guard constraints were infeasible"

        guard_rhs.shape = (len(guard_rhs), ) # flatten into a 1-d array
        assert guard_rhs.shape[0] == guard_csr.shape[0]

        self.guard_csr = guard_csr
        self.guard_rhs = guard_rhs

    def set_reset(self, reset_csr=None, reset_minkowski_csr=None, reset_minkowski_constraints_csr=None,
                  reset_minkowski_constraints_rhs=None):
        '''resets are of the form x' = Rx + My, Cy <= rhs, where y are fresh variables
        the reset_minowski variables can be None if no new variables are needed. If unassigned, the identity
        reset is assumed
 
        x' are the new variables
        x are the old variables       
        reset_csr is R
        reset_minkowski_csr is M
        reset_minkowski_constraints_csr is C
        reset_minkowski_constraints_rhs is rhs
        '''

        assert self.from_mode.a_csr is not None, "A matrix not assigned in predecessor mode {}".format(self.from_mode)

        to_mode_csr = self.to_mode.a_csr

        # reset to an error mode
        if to_mode_csr is None:
            to_mode_csr = self.from_mode.a_csr

        if reset_csr is None:
            assert self.from_mode.a_csr.shape[0] == to_mode_csr.shape[0], "identity reset but num dims changes"
            reset_csr = sp.sparse.identity(self.from_mode.a_csr.shape[0], dtype=float, format='csr')

        if not isinstance(reset_csr, csr_matrix):
            reset_csr = csr_matrix(reset_csr, dtype=float)

        if reset_minkowski_csr is not None and not isinstance(reset_minkowski_csr, csr_matrix):
            reset_minkowski_csr = csr_matrix(reset_minkowski_csr, dtype=float)

        if reset_minkowski_constraints_csr is not None and not isinstance(reset_minkowski_constraints_csr, csr_matrix):
            reset_minkowski_constraints_csr = csr_matrix(reset_minkowski_constraints_csr, dtype=float)

        if reset_minkowski_constraints_rhs is not None and not isinstance(reset_minkowski_constraints_rhs, np.ndarray):
            reset_minkowski_constraints_rhs = np.array(reset_minkowski_constraints_rhs, dtype=float)
        
        assert reset_csr.shape[1] == self.from_mode.a_csr.shape[0], "reset matrix expected {} columns, got {}".format(
            self.from_mode.a_csr.shape[0], reset_csr.shape[1])
        assert reset_csr.shape[0] == to_mode_csr.shape[0], "reset matrix expected {} rows, got {}".format(
            to_mode_csr.shape[0], reset_csr.shape[0])

        if reset_minkowski_constraints_rhs is not None:
            assert len(reset_minkowski_constraints_rhs.shape) == 1
            assert reset_minkowski_constraints_csr is not None
            assert reset_minkowski_constraints_csr.shape[0] == len(reset_minkowski_constraints_rhs)

            new_vars = reset_minkowski_constraints_csr.shape[1]

            if reset_minkowski_csr is None:
                reset_minkowski_csr = sp.sparse.identity(new_vars, dtype=float, format='csr')

            assert isinstance(reset_minkowski_csr, csr_matrix)
            assert reset_minkowski_csr.shape[0] == to_mode_csr.shape[0]
            assert reset_minkowski_csr.shape[1] == new_vars, \
                "expected num reset_minkowski columns({}) to match reset_minkowski_constraints columns({})".format( \
                reset_minkowski_csr.shape[1], new_vars)

            assert lputil.is_feasible(reset_minkowski_constraints_csr, reset_minkowski_constraints_rhs), \
                "reset minkowski variable constraints were infeasible"

        self.reset_csr = reset_csr
        self.reset_minkowski_csr = reset_minkowski_csr
        self.reset_minkowski_constraints_csr = reset_minkowski_constraints_csr
        self.reset_minkowski_constraints_rhs = reset_minkowski_constraints_rhs

    def get_guard_intersection(self, lpi):
        '''do an intersection between this guard and the passed-in lpi
        if there is an intersection, return a new lpi that equals the intersection, otherwise return None
        '''

        # optimized version: first check if every constraint is satisfiable, before checking that they're
        # all satisfied at the same time. The first part can be done quickly by only changing the objective function.
        rv = None
        all_sat = True

        for i, row in enumerate(self.guard_csr):
            # row is csr_matrix of a single row
            
            lpi.set_minimize_direction(row, is_csr=True)

            columns = [lpi.cur_vars_offset + i for i in row.indices]

            result = lpi.minimize(columns=columns, fail_on_unsat=False)

            # sometimes, changing the objective function makes lp infeasible (due to numerical precision issues)
            # this happens on gearbox with small time steps. in this case, just return no transition is possible
            if result is None:
                all_sat = False
                break

            dot_res = np.dot(result, row.data)

            if dot_res > self.guard_rhs[i]:
                all_sat = False
                break

        if all_sat:
            # make the full lpi including all the guard constraints simultaneously
            t_lpi = lpi.clone()

            # add the guard condition
            lputil.add_curtime_constraints(t_lpi, self.guard_csr, self.guard_rhs)

            if t_lpi.is_feasible():
                rv = t_lpi

        return rv

    def __str__(self):
        s = self.from_mode.name + " -> " + self.to_mode.name

        if self.name:
            s += f" ({self.name})"

        return s

class HybridAutomaton(Freezable):
    'The hybrid automaton'

    def __init__(self, name='HybridAutomaton'):
        self.name = name
        self.modes = {} # map name -> mode
        self.transitions = []

        self.freeze_attrs()

    def new_mode(self, name):
        '''add a mode'''

        assert not name in self.modes, "Mode with name '{}' already exists in the automaton".format(name)
        
        m = Mode(self, name, len(self.modes))
        self.modes[m.name] = m
        return m

    def new_transition(self, from_mode, to_mode, name=None):
        '''add a transition'''

        t = Transition(self, from_mode, to_mode, name=name)
        self.transitions.append(t)

        return t

    def check_transitions(self):
        '''
        check that transitios have appropriate resets if the number of variables changes. This is done automatically
        when set_reset is called, but sometimes this may not be called (identity resets). This will check these cases.

        This also checks that guards were assigned and error modes don't have outgoing transitions
        '''

        for t in self.transitions:
            assert t.from_mode.a_csr is not None, \
                "Outgoing transition detected from error mode: {} (not allowed)".format(t)

            assert t.guard_csr is not None, ("Transition '{}' guard was NOT assigned. " + \
                "Use set_guard_true() for always enabled guards, if that's what was intended.").format(t)
            
            if t.to_mode.a_csr is None:
                continue
            
            premode_dims = t.from_mode.a_csr.shape[0]
            postmode_dims = t.to_mode.a_csr.shape[0]

            assert premode_dims == postmode_dims or t.reset_csr is not None, \
                "Transition {} premode has {} dims and postmode has {} dims, but no reset was assigned".format(
                    t, premode_dims, postmode_dims)

    def do_guard_strengthening(self):
        '''
        Strengthen the guards to include the invariants of target modes
        '''

        for t in self.transitions:
            if t.from_mode == t.to_mode or t.time_triggered or t.reset_csr is not None:
                continue

            inv_list = t.to_mode.inv_list
            add_guard_list = []

            for inv_constraint in inv_list:
                already_in_cond_list = False

                for row in range(t.guard_csr.shape[0]):
                    inds = []
                    data = []
                    indptr = [0]

                    for i in range(t.guard_csr.indptr[row], t.guard_csr.indptr[row + 1]):
                        inds.append(t.guard_csr.indices[i])
                        data.append(t.guard_csr.data[i])

                    indptr.append(len(data))
                    csr_vec = csr_matrix((data, inds, indptr), dtype=float, shape=(1, t.guard_csr.shape[1]))
                    guard_constraint = LinearConstraint(csr_vec, t.guard_rhs[row])
                    
                    if guard_constraint.almost_equals(inv_constraint, 1e-13):
                        already_in_cond_list = True
                        break

                if not already_in_cond_list:
                    add_guard_list.append(inv_constraint)

            # add all consrtaints in add_guard_list to the transition's guard (construct a new guard matrix)
            if add_guard_list:
                inds = [n for n in t.guard_csr.indices]
                data = [x for x in t.guard_csr.data]
                indptr = [n for n in t.guard_csr.indptr]
                rhs = [x for x in t.guard_rhs]

                for new_guard_condition in add_guard_list:
                    inds += [n for n in new_guard_condition.csr.indices]
                    data += [x for x in new_guard_condition.csr.data]
                    rhs.append(new_guard_condition.rhs)

                    indptr.append(len(data))

                height = t.guard_csr.shape[0] + len(add_guard_list)
                width = t.guard_csr.shape[1]
                new_guard_csr = csr_matrix((data, inds, indptr), dtype=float, shape=(height, width))
                new_guard_rhs = np.array(rhs, dtype=float)

                t.guard_csr = new_guard_csr
                t.guard_rhs = new_guard_rhs

    def detect_tt_transitions(self, step_size=1.0, num_steps=100, print_func=None):
        '''
        Mark all time-triggered transitions within the automaton.

        This checks for guards where one of the conditions is an inequality involving a single variable, x <= K,
        where the mode invariant has the opposite condition, x >= K (encoded as -x <= -K),
        and the derivative of variable x only depends on a single variable: x' == 5a,
        and the derivative of that variable is all zeros: a' == 0,

        at verification-time, a further check will make sure that a is a constant (not an interval), not equal to 0, 
        and that x is flat (not an interval)
        '''

        # relax constaints by epsilon to prevent missing transition due to floating-point errors
        epsilon = step_size / num_steps / 1000.0

        if print_func is None:
            def simple_print(s):
                'print using the print function'

                print(s)

            print_func = simple_print

        for mode in self.modes.values():
            if mode.a_csr is None: # skip error modes
                continue
            
            # find all derivatives that only depend on constant vars
            tt_vars = get_tt_vars(mode)

            print_func("Checking mode '{}', tt_vars = {}, num transitions = {}".format( \
                mode.name, tt_vars, len(mode.transitions)))

            if not tt_vars:
                continue

            # at this point, we know which variables are constantly changing... check guards / invariants
            for t in mode.transitions:
                if check_and_relax_time_triggered(t, tt_vars, epsilon, print_func):
                    t.time_triggered = True

def get_tt_vars(mode):
    'get all variable indices that only depend on constant terms'

    tt_vars = []

    # find all variables with derivative equal to zero
    constant_vars = []

    for i in range(mode.a_csr.shape[0]):
        nonzeros = mode.a_csr[i].getnnz()

        if nonzeros == 0: # row of all zeros
            constant_vars.append(i)

    for row_index, row in enumerate(mode.a_csr):
        if row.getnnz() == 0: # skip constant variables
            continue

        all_constant = True

        for i, col in enumerate(row.indices):
            if row.data[i] != 0 and not col in constant_vars:
                all_constant = False
                break

        if all_constant:
            tt_vars.append(row_index)

    return tt_vars

def check_and_relax_time_triggered(t, tt_vars, epsilon, print_func):
    '''is the passed-in transition time triggered? If so, relax it by epsilon

    t is a transition
    tt_vars is a list of variables with derivatives that only depend on constant terms, like 't' in: t' = a, a' = 0

    '''

    rv = False

    print_func("checking transition {}, t.guard.rhs = {}".format(t, t.guard_rhs))

    # all of the conditions in the invariant must involve tt_vars and
    # one of the conditions in the guard must be the inverse of an invariant condition

    all_invariant_tt = True

    for lc in t.from_mode.inv_list:
        print_func("checking invariant is tt: {} <= {}".format(lc.csr.toarray(), lc.rhs))

        for i, col in enumerate(lc.csr.indices):
            if lc.csr.data[i] != 0 and not col in tt_vars:
                all_invariant_tt = False
                break

    if all_invariant_tt:
        print_func("all invariant conditions are on time-triggered variables... checking for tt guard")

        for index, (row, rhs) in enumerate(zip(t.guard_csr, t.guard_rhs)):
            print_func("checking guard {} <= {}".format(row.toarray(), rhs))
                                
            all_tt_vars = True
            
            for i, col in enumerate(row.indices):
                if row.data[i] != 0 and not col in tt_vars:
                    all_tt_vars = False
                    break

            print_func("t.guard_csr = {}, all_tt_vars = {}".format(row.toarray(), all_tt_vars))

            if all_tt_vars:
                # check if there is a mode invariant with the opposite condition
                found_invariant = False

                for lc in t.from_mode.inv_list:
                    print_func("checking for inverse with invariant {} <= {}".format(lc.csr.toarray(), lc.rhs))

                    if lc.rhs == -1 * rhs and (-1 * lc.csr != row[0]).nnz == 0:
                        print_func("found opposite invariant condition!")
                        found_invariant = True
                        break

                if found_invariant:
                    t.guard_rhs[index] += epsilon # perform epsilon relaxation
                    print_func("relaxed guard to {} <= {}".format(row.toarray(), t.guard_rhs[index]))
                
                    rv = True

            print_func("Guard {} <= {} {} time triggered".format(row.toarray(), rhs, "was" if rv else "was NOT"))
                

        print_func("Static Check: Transition {} {} time triggered".format(t, "is" if rv else "is NOT"))
                   
    return rv

def was_tt_taken(state_lpi, t, step_size, num_steps):
    '''do the run-time checks to see if this transition was a time-triggred one

    if we have x <= K, with x' = -5 * a,
    make sure that a is a constant (not an interval), not equal to 0, and that x is flat (not an interval)
    '''

    # epsilon tolerance is computed to be fatter than the relaxation performed for tt transition
    epsilon = 2 * step_size / num_steps / 1000.0
    #print(f". epsilon in was_tt_taken: {epsilon}")

    rv = False
    tt_vars = get_tt_vars(t.from_mode)

    #print(f". tt_vars: {tt_vars}")

    # find the tt-guard condition first
    for row, rhs in zip(t.guard_csr, t.guard_rhs):
        #print(".checking guard {} <= {}".format(row.toarray(), rhs))

        all_tt_vars = True
        for i, col in enumerate(row.indices):
            if row.data[i] != 0 and not col in tt_vars:
                all_tt_vars = False
                break

        #print(". t.guard_csr = {}, all_tt_vars = {}".format(row.toarray(), all_tt_vars))
        found_invariant = False

        if all_tt_vars:
            # check if there is a mode invariant with the opposite condition

            for lc in t.from_mode.inv_list:
                #print(". checking for inverse with invariant {} <= {}".format(lc.csr.toarray(), lc.rhs))

                if abs(lc.rhs - (-1 * rhs)) < epsilon and (-1 * lc.csr != row[0]).nnz == 0:
                    #print(". found opposite invariant condition")
                    found_invariant = True
                    break
                
        if not found_invariant: # this was not the time-triggered condition in the current guard condition
            continue

        dims = state_lpi.dims

        # first look at the guard condition, and see which variables it uses (find 'x' in x <= K, x' == a)
        guard_vars = []

        for i, col in enumerate(row.indices):
            if row.data[i] != 0:
                guard_vars.append(col)

        # next, find all the 'a' in x' == a
        affine_vars = {}
        a_csr = t.from_mode.a_csr

        for var in guard_vars:
            for index in range(a_csr.indptr[var], a_csr.indptr[var+1]):
                if a_csr.data[index] != 0:
                    affine_vars[a_csr.indices[index]] = True # insert it into the dict

        #print(". guard vars = {}, affine_vars = {}".format(guard_vars, affine_vars.keys()))

        all_affine_nonzero = True

        for var in affine_vars:
            min_dir = [1 if n == var else 0 for n in range(dims)]
            max_dir = [-1 if n == var else 0 for n in range(dims)]
            col = state_lpi.cur_vars_offset + var

            min_val = state_lpi.minimize(direction_vec=min_dir, columns=[col])[0]
            max_val = state_lpi.minimize(direction_vec=max_dir, columns=[col])[0]

            #print(". affine var {}, min_val = {}, max_val = {}".format(var, min_val, max_val))

            if abs(max_val - min_val) > epsilon or abs(max_val) < epsilon:
                all_affine_nonzero = False
                break

        if all_affine_nonzero:
            # make sure x is flat... first find x
            #print(". all affine nonzero, optimizing guard direction")

            state_cols = [state_lpi.cur_vars_offset + n for n in range(dims)]
            direction = row.toarray()[0]

            min_state = state_lpi.minimize(direction_vec=direction, columns=state_cols)
            max_state = state_lpi.minimize(direction_vec=-1 * direction, columns=state_cols)

            min_val = np.dot(min_state, direction)
            max_val = np.dot(max_state, direction)

            #print(". min_val = {}, max_val = {}, rhs = {}".format(min_val, max_val, rhs))

            if abs(max_val - min_val) < epsilon and min_val <= rhs:
                rv = True
                
    return rv



'''
Hylaa Settings File
Stanley Bak, 2018
'''

import math

from matplotlib import animation

from hylaa.util import Freezable
from hylaa import aggstrat

# Force matplotlib to not use any Xwindows backend.
#import matplotlib
#matplotlib.use('Agg')

# Suppress floating point printing overflow/underflow printing in numpy
#import numpy as np
#np.set_printoptions(suppress=True)

class HylaaSettings(Freezable):  # pylint: disable=too-few-public-methods
    'Settings for the computation'

    STDOUT_NONE, STDOUT_NORMAL, STDOUT_VERBOSE, STDOUT_DEBUG = range(4)

    # Approximation Models: None: discrete-time, chull: convex hull only (different simulation semantics),
    #                       lgg: support function method from Le Guernic'10
    APPROX_NONE, APPROX_CHULL, APPROX_LGG = range(3)

    def __init__(self, step_size, max_time):
        plot_settings = PlotSettings()

        self.step_size = step_size # simulation step size
        self.num_steps = int(math.ceil(max_time / step_size))

        self.plot = plot_settings
        self.stdout = HylaaSettings.STDOUT_NORMAL
        self.stdout_colors = [None, "white", "blue", "yellow"] # colors for each level of printing

        ### SIMULATION-EQUIVALENT SEMANTICS / COMPUTATION PARAMETERS ###
        self.process_urgent_guards = False # allow zero continuous-post steps between transitions?
        self.do_guard_strengthening = True # add invariants of target modes to each guard?
        self.optimize_tt_transitions = True # auto-detect time-triggered transitions and use single-step semantics?
        self.approx_model = HylaaSettings.APPROX_NONE
        self.skip_zero_dynamics_modes = True

        # what to do when an error appears reachable
        self.stop_on_aggregated_error = False # stop whenever any state (aggregated or not) reaches an error mode
        self.stop_on_concrete_error = True # stop whenver a concrete state reaches an error
        self.make_counterexample = True # save counter-example to data structure / file?
        
        self.aggstrat = aggstrat.Aggregated() # aggregation strategy class

        # for deterministic random numbers (simulations / color selection)
        self.random_seed = 0

        self.freeze_attrs()

class PlotSettings(Freezable): # pylint: disable=too-few-public-methods,too-many-instance-attributes
    'plot settings container'

    PLOT_NONE = 0 # don't plot (safety checking only; for performance measurement)
    PLOT_LIVE = 1 # plot the computation live as we're computing
    PLOT_INTERACTIVE = 2 # live plotting with pauses and buttons upon certain events
    PLOT_IMAGE = 3 # save the image plot to a file
    PLOT_VIDEO = 4 # save a video to a file

    def __init__(self):
        self.plot_mode = PlotSettings.PLOT_NONE
        
        self.store_plot_result = False # store the reachable plot data inside the computation result object?

        self.filename = None # filename to print data to for certain plot modes

        self.plot_size = (8, 8) # inches

        # these settings can be lists, in which case we'll make multiple plots
        self.xdim_dir = 0 # plotting x dimension number, direction (np.array), None (time), or dict: mode_name -> dir
        self.ydim_dir = 1 # plotting y dimension number, direction (np.array), None (time), or dict: mode_name -> dir
        self.label = LabelSettings() # plot title, axis labels, font sizes, ect.
        self.extra_collections = None # list of extra animation collections

        self.num_angles = 512 # how many evenly-spaced angles to put into plot_vecs

        self.draw_stride = 1 # draw every n frames (good to inscrease if number of steps is large)
        
        self.reachable_poly_width = 2 # width of reachable polygon outlines
        self.extend_plot_range_ratio = 0.1 # extend plot axis range 10% at a time

        self.sim_line_color = 'black' # simulation line color
        self.sim_line_width = 0.2 # width of simulation lines

        self.grid = True
        self.grid_xtics = None # override default xtics value, for example np.linspace(0.0, 5.0, 1.0)
        self.grid_ytics = None # override default ytics value, for example np.linspace(0.0, 5.0, 1.0)

        self.use_markers_for_small = True # draw markers when the reachable set is tiny instead of (invisible) polygons

        self.show_counterexample = True # draw concrete counter-example in the last frame?

        self.interactive_skip_count = 0 # when using PLOT_INTERACTIVE, auto-click 'next' this many times

        self.video_fps = 40
        self.video_extra_frames = 40 # extra frames at the end of a video so it doesn't end so abruptly
        self.video_pause_frames = 20 # frames to render in video whenever a 'pause' occurs
        self.video_show_frame = True # show the frame counter?

        # function which returns the Writer with the desired settings used to create a video, used for video export
        def make_video_writer():
            'returns the Writer to create a video for export'

            writer_class = animation.writers['ffmpeg']
            return writer_class(fps=self.video_fps, metadata=dict(artist='Me'), bitrate=1800)

        self.make_video_writer_func = make_video_writer

        self.freeze_attrs()

class LabelSettings(Freezable):
    'settings for labels such as plot title, plot font size, ect.'

    def __init__(self):
        self.x_label = None
        self.y_label = None
        self.title = None

        self.title_size = 32
        self.label_size = 24
        self.tick_label_size = 18
        self.axes_limits = None # fixed axes limits; a 4-tuple (xmin, xmax, ymin, ymax) or None for auto

        self.freeze_attrs()

    def big(self, size=30):
        'increase sizes of labels'

        self.title_size = size
        self.label_size = size
        self.tick_label_size = int(0.8 * size)

    def turn_off(self):
        'turn off plot labels'

        self.x_label = ''
        self.y_label = ''
        self.title = ''



'''
Generalized Star and other Star data structures
Stanley Bak
Aug 2016
'''

import math

import numpy as np
import scipy as sp

from matplotlib.path import Path

from hylaa import lpplot, lputil

from hylaa.hybrid_automaton import Mode
from hylaa.timerutil import Timers
from hylaa.util import Freezable
from hylaa.lpinstance import LpInstance
from hylaa.settings import HylaaSettings

class StateSet(Freezable):
    '''
    A set of states (possibly aggregated) in the same mode.
    '''

    def __init__(self, lpi, mode, cur_steps_since_start=None, aggdag_op_list=None, is_concrete=True):
        assert isinstance(lpi, LpInstance)
        assert isinstance(mode, Mode)

        self.mode = mode
        self.lpi = lpi
    
        self.cur_step_in_mode = 0

        if cur_steps_since_start is not None:
            assert len(cur_steps_since_start) == 2 # needs to be an interval in case this stateset is an aggregation
            self.cur_steps_since_start = cur_steps_since_start.copy()
        else:
            self.cur_steps_since_start = [0, 0]

        self.is_concrete = is_concrete

        if aggdag_op_list is None: # assume its an initial stateset
            aggdag_op_list = [None]

        assert isinstance(aggdag_op_list, (list, tuple))        
        self.aggdag_op_list = aggdag_op_list # list of OpTransition objects that created this state set

        # the LP row of the strongest constraint for each invariant condition
        # this is used to eliminate redundant constraints as the lpi is intersected with the invariant at each step
        self.invariant_constraint_rows = [None] * len(self.mode.inv_list)

        # mode might be an error mode, in which case a_csr is None
        self.basis_matrix = None if mode.a_csr is None else np.identity(mode.a_csr.shape[0])

        if mode.a_csr is not None and not np.allclose(lputil.get_basis_matrix(lpi), self.basis_matrix):
            raise RuntimeError("lpi basis matrix in StateSet constructor was not the identity. " + \
                               "Did you construct the lpi using the lputil.from_*() functions?")
        
        self.input_effects_list = None if mode.b_csr is None else [] # list of input effects at each step

        self.aggstring = None # aggstring that led to this state, like 'full', or '010'

        # approximation model variables
        self.lgg_beta = None

        #### plotting variables below ####
        self._verts = None # cached vertices at the current step
        self.assigned_plot_dim = False # set to True on first call to verts()
        self.xdim = None # set on first call to verts()
        self.ydim = None # set on first call to verts()

        # map of step number in mode -> list (for each subplot) of pairs: (matploatlib Path object list, index)
        self.step_to_paths = {} 

        self.freeze_attrs()

    def __str__(self):
        'short string representation of this state set'

        return f"[StateSet, mode:{self.mode.name} @ step {self.cur_step_in_mode}, {self.get_full_aggstring()}]"

    def get_full_aggstring(self):
        'get the full aggstring that led to this stateset'

        rv = None

        if self.aggdag_op_list[0] is None:
            assert self.aggstring == 'full'
            rv = 'init'
        else:
            op = self.aggdag_op_list[0]
            parent = op.parent_node.stateset
            parent_aggstring = parent.get_full_aggstring()
            tindex = parent.mode.transitions.index(op.transition)

            rv = parent_aggstring + f":t{tindex}_" + self.aggstring

        return rv

    def step(self, step_in_mode=None):
        '''update the star based on values from a new simulation time instant

        the default is to advance by one step, otherwise step_in_mode can force
        going to a specific step number
        '''

        Timers.tic("step")

        if step_in_mode is None:
            step_in_mode = self.cur_step_in_mode + 1

        num_steps = step_in_mode - self.cur_step_in_mode

        # we can't do negative steps because we add input effects in the lpi for each step
        assert num_steps >= 0, "step() called with negative num steps (mode: " + \
          f"{self.mode.name}, cur_step_in_mode: {self.cur_step_in_mode}, requested_step: {step_in_mode})"

        if num_steps > 0:
            Timers.tic('get_bm')
            self.basis_matrix, input_effects_matrix = self.mode.time_elapse.get_basis_matrix(step_in_mode)
            Timers.toc('get_bm')

            Timers.tic('set_bm')
            lputil.set_basis_matrix(self.lpi, self.basis_matrix)
            Timers.toc('set_bm')

            if input_effects_matrix is not None:
                Timers.tic('input effects matrix')
                # if we're doing multiple steps here we need to get each step's input effects matrix
                for step in range(self.cur_step_in_mode + 1, step_in_mode):
                    _, ie_mat = self.mode.time_elapse.get_basis_matrix(step)
                    self.input_effects_list.append(ie_mat)
                    lputil.add_input_effects_matrix(self.lpi, ie_mat, self.mode, self.lgg_beta)

                # add the input effects matrix for the final step (computed before with basis matrix)
                self.input_effects_list.append(input_effects_matrix)
                lputil.add_input_effects_matrix(self.lpi, input_effects_matrix, self.mode, self.lgg_beta)
                Timers.toc('input effects matrix')

                #print(f".ss lp columns = {self.lpi.get_num_cols()}")

            self.cur_step_in_mode += num_steps
            self.cur_steps_since_start[0] += num_steps
            self.cur_steps_since_start[1] += num_steps
            self._verts = None # cached vertices no longer valid

        Timers.toc("step")

    def verts(self, plotman, subplot=0):
        'get the vertices for plotting this state set, wraps around so rv[0] == rv[-1]'

        Timers.tic('verts')

        if self._verts is None:
            self._verts = [None] * plotman.num_subplots

        if self._verts[subplot] is None:
            min_time = self.cur_steps_since_start[0] * plotman.core.settings.step_size
            max_time = self.cur_steps_since_start[1] * plotman.core.settings.step_size
            time_interval = (min_time, max_time)

            if not self.assigned_plot_dim:
                self.assigned_plot_dim = True

                self.xdim = []
                self.ydim = []

                for i in range(plotman.num_subplots):
                    self.xdim.append(plotman.settings.xdim_dir[i])
                    self.ydim.append(plotman.settings.ydim_dir[i])

                    if isinstance(self.xdim[i], dict):
                        assert self.mode.name in self.xdim[i], "mode {} not in xdim plot direction dict".format(
                            self.mode.name)
                        self.xdim[i] = self.xdim[i][self.mode.name]

                    if isinstance(self.ydim[i], dict):
                        assert self.mode.name in self.ydim[i], "mode {} not in ydim plot direction dict".format(
                            self.mode.name)
                        self.ydim[i] = self.ydim[i][self.mode.name]

            self._verts[subplot] = lpplot.get_verts(self.lpi, xdim=self.xdim[subplot], ydim=self.ydim[subplot], \
                                           plot_vecs=plotman.plot_vec_list[subplot], cur_time=time_interval)
            
            assert self._verts[subplot] is not None, "verts() was unsat"
            
        Timers.toc('verts')

        return self._verts[subplot]

    def set_plot_path(self, subplot, path_list, path_index):
        '''
        set the matplotlib path object for this stateset at the current step
        '''

        step = self.cur_step_in_mode

        if step not in self.step_to_paths:
            l = []
            self.step_to_paths[step] = l
        else:
            l = self.step_to_paths[step]

        while len(l) <= subplot:
            l.append([])

        l[subplot] = (path_list, path_index)

    def del_plot_path(self, step):
        '''
        delete a plotted matplotlib Path object for this stateset.

        returns a list of verts (one for each subplot) that was deleted, or None if stateset was
        already deleted (can happen with recursive deaggregation)
        '''

        rv = []

        if not step in self.step_to_paths:
            rv = None
        else:
            l = self.step_to_paths.pop(step) # removes it from step_to_paths as well

            codes = [Path.MOVETO, Path.CLOSEPOLY]
            verts = [(0, 0), (0, 0)]

            # for every subplot
            for path_list, index in l:
                rv.append(path_list[index].vertices)

                # this is what actually removes it from the plot
                path_list[index] = Path(verts, codes)

        return rv

    def apply_approx_chull(self):
        '''
        apply convex hull approximation model
        '''

        lpi_one_step = self.lpi.clone()

        Timers.tic('get_bm')
        bm, ie_mat = self.mode.time_elapse.get_basis_matrix(1)
        Timers.toc('get_bm')

        Timers.tic('set_bm')
        lputil.set_basis_matrix(lpi_one_step, bm)
        Timers.toc('set_bm')

        if ie_mat is not None:
            Timers.tic('input effects matrix')
            lputil.add_input_effects_matrix(lpi_one_step, ie_mat, self.mode)
            Timers.toc('input effects matrix')

        lpi_list = [self.lpi, lpi_one_step]
        self.lpi = lputil.aggregate_chull(lpi_list, self.mode)

    def apply_approx_lgg(self):
        '''
        apply lgg approximation model from equation (2) in Lemma 1 of:
        "Reachability analysis of linear systems using support functions",
        Le Guernic, C., Girard, A., Nonlinear Analysis: Hybrid Systems, 2010
        '''

        has_inputs = self.mode.b_csr is not None

        # use infinity norm
        a_norm = sp.sparse.linalg.norm(self.mode.a_csr, ord=np.inf)

        lpi_one_step = self.lpi.clone()

        Timers.tic('get_bm')
        bm, _ = self.mode.time_elapse.get_basis_matrix(1)
        Timers.toc('get_bm')

        Timers.tic('set_bm')
        lputil.set_basis_matrix(lpi_one_step, bm)
        Timers.toc('set_bm')

        mode = self.mode

        tau = mode.time_elapse.step_size
        r_x0 = lputil.compute_radius_inf(self.lpi)

        if has_inputs:
            v_set = lputil.from_input_constraints(mode.b_csr, mode.u_constraints_csc, mode.u_constraints_rhs, mode)
            r_v = lputil.compute_radius_inf(v_set)
            
            # minkowski sum with tau * V
            # V is the input set, V = B U
            lputil.scale_with_bm(v_set, tau)
            msum = lputil.minkowski_sum([lpi_one_step, v_set], mode)
        else:
            r_v = 0
            msum = lpi_one_step

        tol = 1e-7

        if a_norm < tol:
            print(f"Warning: norm of dynamics A matrix was small ({a_norm}), using alpha = 0 and " +
                                  "beta = 0 in LGG approximation model")
            alpha = 0
        else:
            alpha = (math.exp(tau * a_norm) - 1 - tau * a_norm) * (r_x0 + r_v/a_norm)

        #print(f".ss alpha={alpha}")

        if alpha != 0:
            # bloat by alpha (minkowski sum)
            lputil.bloat(msum, alpha)

        self.lpi = lputil.aggregate_chull([self.lpi, msum], mode)

        if a_norm > tol and has_inputs:
            # precompute beta as well
            self.lgg_beta = (math.exp(tau * a_norm) - 1 - tau * a_norm) * (r_v/a_norm)
            #print(f".ss beta={self.lgg_beta}")

            assert self.lgg_beta > tol, f"lgg approx model beta was too close to zero: {self.lgg_beta}"
            assert self.lgg_beta < 1e5, f"lgg approx model beta was too large (use a smaller step): {self.lgg_beta}"

            self.lpi.set_minimize_direction([-1] * self.lpi.dims)

            self.mode.time_elapse.use_lgg_approx()

    def apply_approx_model(self, approx_model):
        '''
        apply the approximation model to bloat the current (initial) set of states

        approx_model - one of the APPROX_ values defined in HylaaSettings
        '''

        assert self.cur_step_in_mode == 0, "approximation model should be applied before any continuous post operations"
        assert self.mode.time_elapse is not None, "init_time_elapse() must be called before apply_approx_model()"

        if approx_model == HylaaSettings.APPROX_CHULL:
            self.apply_approx_chull()
        elif approx_model == HylaaSettings.APPROX_LGG:
            self.apply_approx_lgg()
        elif approx_model != HylaaSettings.APPROX_NONE:
            assert f"Unknown approx_model from settings: {approx_model}"



'''
Tests for Hylaa aggregation. Made for use with py.test
'''

import math
import random

import matplotlib.pyplot as plt

import numpy as np
from scipy.sparse import csr_matrix
from scipy.linalg import expm

from hylaa.hybrid_automaton import HybridAutomaton
from hylaa.settings import HylaaSettings, PlotSettings
from hylaa.core import Core
from hylaa.stateset import StateSet
from hylaa import lputil, lpplot
from hylaa.aggdag import OpTransition, AggDagNode
from hylaa.aggstrat import Aggregated

from util import pair_almost_in, assert_verts_is_box

def test_box_aggregate2():
    'tests box aggregation'

    mode = HybridAutomaton().new_mode('mode_name')

    lpi1 = lputil.from_box([[0, 1], [0, 1]], mode)
    lpi2 = lputil.from_box([[1, 2], [1, 2]], mode)

    agg_dirs = np.identity(2)

    # box aggregation
    lpi = lputil.aggregate([lpi1, lpi2], agg_dirs, mode)

    verts = lpplot.get_verts(lpi)
    assert_verts_is_box(verts, [[0, 2], [0, 2]])

    # test setting basis matrix after aggregation
    lputil.set_basis_matrix(lpi, np.identity(2))

    verts = lpplot.get_verts(lpi)
    assert_verts_is_box(verts, [[0, 2], [0, 2]])

    lputil.set_basis_matrix(lpi, -1 * np.identity(2))

    verts = lpplot.get_verts(lpi)
    assert_verts_is_box(verts, [[-2, 0], [-2, 0]])

def test_rotated_aggregate():
    'tests rotated aggregation'

    mode = HybridAutomaton().new_mode('mode_name')
    lpi1 = lputil.from_box([[0, 1], [0, 1]], mode)
    lpi2 = lputil.from_box([[1, 2], [1, 2]], mode)

    sq2 = math.sqrt(2) / 2.0

    agg_dirs = np.array([[sq2, sq2], [sq2, -sq2]], dtype=float)

    lpi = lputil.aggregate([lpi1, lpi2], agg_dirs, mode)

    assert lputil.is_point_in_lpi([0, 0], lpi)
    assert lputil.is_point_in_lpi([2, 2], lpi)
    assert lputil.is_point_in_lpi([1, 2], lpi)
    assert lputil.is_point_in_lpi([2, 1], lpi)
    assert lputil.is_point_in_lpi([0, 1], lpi)
    assert lputil.is_point_in_lpi([1, 0], lpi)

    verts = lpplot.get_verts(lpi)

    assert len(verts) == 5

    for p in [(0.5, -0.5), (-0.5, 0.5), (2.5, 1.5), (1.5, 2.5)]:
        assert pair_almost_in(p, verts)

    assert verts[0] == verts[-1]


def test_box_aggregate3():
    'tests box aggregation with 3 boxes'

    mode = HybridAutomaton().new_mode('mode_name')
    
    lpi1 = lputil.from_box([[-2, -1], [-0.5, 0.5]], mode)
    lpi2 = lpi1.clone()
    lpi3 = lpi1.clone()

    basis2 = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi2, basis2)

    basis3 = np.array([[-1, 0], [0, -1]], dtype=float)
    lputil.set_basis_matrix(lpi3, basis3)

    plot_vecs = lpplot.make_plot_vecs(256, offset=0.1) # use an offset to prevent LP dir from being aligned with axis

    # bounds for lpi1 should be [[-2, -1], [-0.5, 0.5]]
    verts = lpplot.get_verts(lpi1, plot_vecs=plot_vecs)
    assert_verts_is_box(verts, [[-2, -1], [-0.5, 0.5]])

    # bounds for lpi2 should be [[-0.5, 0.5], [1, 2]]
    verts = lpplot.get_verts(lpi2, plot_vecs=plot_vecs)
    assert_verts_is_box(verts, [[-0.5, 0.5], [1, 2]])

    # bounds for lpi3 should be [[2, 1], [-0.5, 0.5]]
    verts = lpplot.get_verts(lpi3, plot_vecs=plot_vecs)
    assert_verts_is_box(verts, [[2, 1], [-0.5, 0.5]])
 
    # box aggregation, bounds should be [[-2, 2], [-0.5, 2]]
    agg_dirs = np.identity(2)
    lpi = lputil.aggregate([lpi1, lpi2, lpi3], agg_dirs, mode)

    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)
    assert_verts_is_box(verts, [[-2, 2], [-0.5, 2]])

def test_aggregate_on_subspace():
    '''
    test aggregation when the dynamics and sets are only on a subspace. 
    '''

    # dynamics are x' == 1, y' == 0, a' == 0
    # lpi1 is [0, 1] x [0, 1] x [1, 1]
    # lpi2 is [3, 4] x [0, 1] x [1, 1]

    # aggregation shouldn't need to introduce a variable along the y direction

    mode = HybridAutomaton().new_mode('mode_name')
    lpi1 = lputil.from_box([[0, 1], [0, 1], [1, 1]], mode)
    lpi2 = lputil.from_box([[4, 5], [0, 1], [1, 1]], mode)

    #a_csr = csr_matrix(np.array([[0, 0, 1], [0, 0, 0], [0, 0, 0]], dtype=float))
    sqr = math.sqrt(2) / 2
    agg_dirs = np.array([[1, 0, 0], [0, sqr, sqr], [0, sqr, -sqr]], dtype=float)

    # box aggregation
    lpi = lputil.aggregate([lpi1, lpi2], agg_dirs, mode)

    # lpi1 corners
    for pt in [(0, 0, 1), (0, 1, 1), (1, 0, 1), (1, 1, 1)]:
        assert lputil.is_point_in_lpi(pt, lpi)

    # lpi2 corners
    for pt in [(4, 0, 1), (4, 1, 1), (5, 0, 1), (5, 1, 1)]:
        assert lputil.is_point_in_lpi(pt, lpi)

    # make sure we have new variable names
    names = lpi.get_names()

    expected_names = ["m0_i0", "m0_i1", "m0_i2", "m0_c0", "m0_c1", "m0_c2"]

    assert names == expected_names

def test_aggregate_self():
    '''
    test aggregation on an identical set.
    '''

    mode = HybridAutomaton().new_mode('mode_name')
    lpi1 = lputil.from_box([[-2, -1], [-10, 20], [100, 200]], mode)
    lpi2 = lputil.from_box([[-2, -1], [-10, 20], [100, 200]], mode)

    agg_dirs = np.identity(3)

    # box aggregation
    lpi = lputil.aggregate([lpi1, lpi2], agg_dirs, mode)
    
    assert lpi.is_feasible()

    verts = lpplot.get_verts(lpi, xdim=0, ydim=1)
    assert_verts_is_box(verts, [[-2, -1], [-10, 20]])

    verts = lpplot.get_verts(lpi, xdim=0, ydim=2)
    assert_verts_is_box(verts, [[-2, -1], [100, 200]])

    # make sure no extra variables in lp
    names = lpi.get_names()

    expected_names = ["m0_i0", "m0_i1", "m0_i2", "m0_c0", "m0_c1", "m0_c2"]

    assert names == expected_names

    assert lpi.get_num_rows() == 3 + 3*2

def test_reorthogonalize_matrix():
    'tests the reorthgonalize_matrix function'

    mat = np.array([[1, 0, 0], [1, 1, 0], [1, 1, 0.5]], dtype=float)
    out = lputil.reorthogonalize_matrix(mat, 3)
    expected = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)
    assert np.allclose(out, expected)

    mat = np.array([[1, 0, 0], [1, 0, 0], [0, 0, 0.5]], dtype=float)
    out = lputil.reorthogonalize_matrix(mat, 3)
    expected = np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=float)
    assert np.allclose(out, expected)

    sqr = math.sqrt(2) / 2
    mat = np.array([[1, 1], [0, 0], [2, 2]], dtype=float)
    out = lputil.reorthogonalize_matrix(mat, 2)
    assert np.allclose(out[0], np.array([sqr, sqr], dtype=float))

    mat = np.array([[1, 1, 0, 0, 0], [0, 0, 2, 0, 0], [2, 1, 2.5, 0, 0]], dtype=float)
    out = lputil.reorthogonalize_matrix(mat, 5)

    assert out.shape == (5, 5)
    assert np.allclose(out[0], np.array([sqr, sqr, 0, 0, 0], dtype=float))
    assert np.allclose(out[1], np.array([0, 0, 1, 0, 0], dtype=float))
    assert np.allclose(out[2][2:], np.array([0, 0, 0], dtype=float))

    for a, row_a in enumerate(out):
        assert abs(np.linalg.norm(row_a) - 1.0) < 1e-6, "rows should be normalized"

        for b, row_b in enumerate(out):
            if a == b:
                continue

            assert np.dot(row_a, row_b) < 1e-6, "rows should be orthononal"

def test_aggregate3():
    'tests aggregation of 3 sets, inspired by the harmonic oscillator system'

    mode = HybridAutomaton().new_mode('mode_name')
    lpi1 = lputil.from_box([[0, 1], [0, 1]], mode)

    # middle set is a diamond
    mat = [[1, 1], [-1, -1], [1, -1], [-1, 1]]
    s = 3.5
    rhs = [6+s, -(6-s), s, s]
    lpi2 = lputil.from_constraints(mat, rhs, mode)
    
    lpi3 = lputil.from_box([[5, 6], [5, 6]], mode)

    lpi_list = [lpi1, lpi2, lpi3]
    verts = []

    for lpi in lpi_list:
        verts += lpplot.get_verts(lpi)

        #xs, ys = zip(*lpplot.get_verts(lpi))
        #plt.plot(xs, ys, 'k-')

    random.seed(0)

    for _ in range(10):
        random_mat = np.random.rand(2, 2)
        agg_dirs = lputil.reorthogonalize_matrix(random_mat, 2)

        lpi = lputil.aggregate(lpi_list, agg_dirs, mode)

        #xs, ys = zip(*lpplot.get_verts(lpi))
        #plt.plot(xs, ys, 'r--')

        for vert in verts:
            assert lputil.is_point_in_lpi(vert, lpi)

    #plt.show()

def test_chull_ha5():
    'test convex hull aggregation of harmonic oscillator with 5 sets'

    mode = HybridAutomaton().new_mode('mode_name')

    steps = 5
    step_size = math.pi/4

    lpi_list = []

    a_mat = np.array([[0, 1], [-1, 0]], dtype=float)

    for step_num in range(steps):
        box = [[-5, -4], [-0.5, 0.5]]
        
        lpi = lputil.from_box(box, mode)

        t = step_num * step_size
        basis_mat = expm(a_mat * t)
        lputil.set_basis_matrix(lpi, basis_mat)

        lpi_list.append(lpi)
        
    verts = []

    for lpi in lpi_list:
        verts += lpplot.get_verts(lpi)

        xs, ys = zip(*lpplot.get_verts(lpi))
        #plt.plot(xs, ys, 'k-')

    lpi = lputil.aggregate_chull(lpi_list, mode)
    #xs, ys = zip(*lpplot.get_verts(lpi))
    #plt.plot(xs, ys, 'r--')

    #plt.show()

    # test if it's really convex hull
    assert lputil.is_point_in_lpi([0, 4.5], lpi)

    for vert in verts:
        assert lputil.is_point_in_lpi(vert, lpi)

def test_chull():
    'tests aggregation of a cirle of sets using convex hull'

    mode = HybridAutomaton().new_mode('mode_name')

    r = 1.0
    eps = 0.05
    num_sets = 16
    lpi_list = []

    for theta in np.linspace(0, 2*math.pi, num_sets, endpoint=False):
        y = r * math.sin(theta)
        x = r * math.cos(theta)
        
        mat = [[1, 0], [-1, 0], [0, 1], [0, -1]]
        rhs = [x + eps, -(x - eps), y + eps, -(y - eps)]
        lpi = lputil.from_constraints(mat, rhs, mode)
    
        lpi_list.append(lpi)
        
    #verts = []

    #for lpi in lpi_list:
    #    verts += lpplot.get_verts(lpi)

    #    xs, ys = zip(*lpplot.get_verts(lpi))
    #    plt.plot(xs, ys, 'k-')

    lpi = lputil.aggregate_chull(lpi_list, mode)

    #xs, ys = zip(*lpplot.get_verts(lpi))
    #plt.plot(xs, ys, 'r--')

    #for vert in verts:
    #    assert lputil.is_point_in_lpi(vert, lpi)

    #plt.show()

    # test if it's really convex hull
    for theta in np.linspace(0, 2*math.pi, num_sets):
        y = (r + 2*eps) * math.sin(theta)
        x = (r + 2*eps) * math.cos(theta)
        
        assert not lputil.is_point_in_lpi([x, y], lpi)

        y = (r - 2*eps) * math.sin(theta)
        x = (r - 2*eps) * math.cos(theta)

        assert lputil.is_point_in_lpi([x, y], lpi)
        
def test_plain():
    'test plain aggregation of states across discrete transitions'

    # m1 dynamics: x' == 1, y' == 0, x0, y0: [0, 1], step: 1.0
    # m1 invariant: x <= 3
    # m1 -> m2 guard: True
    # m2 dynamics: x' == 0, y' == 1
    # time bound: 4
    # excepted final states to be: x: [0, 4], y: [4,5]
    # x is [1, 4] because no transitions are allowed at step 0 (simulation-equiv semantics) and a transition is
    #        allowed one step after the invariant becomes false
    # y is [4,5] because after aggregation, the time elapsed for the aggregated set will be 0.0, the minimum

    ha = HybridAutomaton()

    # mode one: x' = 1, y' = 0, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 0], [0, 0, 0]])

    # mode two: x' = 0, y' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, 0], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 3.0
    m1.set_invariant([[1, 0, 0]], [3.0])

    # guard: True
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard(csr_matrix((0, 0)), [])

    # initial set has x0 = [0, 1], t = [0, 1], a = 1
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 4.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.store_plot_result = True

    core = Core(ha, settings)
    result = core.run(init_list)

    # check history
    state = result.last_cur_state

    assert state.mode == m2
    assert len(state.aggdag_op_list) > 1
    
    op0 = state.aggdag_op_list[0]
    op1 = state.aggdag_op_list[1]
    assert isinstance(op0, OpTransition)

    assert len(core.aggdag.roots) == 1
   
    assert op0.child_node.stateset.mode is m2
    assert op0.transition == trans1
    assert op0.parent_node == core.aggdag.roots[0]
    assert isinstance(op0.poststate, StateSet)
    assert op0.step == 1
    assert isinstance(op0.child_node, AggDagNode)
    assert op0.child_node == op1.child_node
    assert op0.child_node not in core.aggdag.roots

    assert len(op0.parent_node.stateset.aggdag_op_list) == 1
    assert op0.parent_node.stateset.aggdag_op_list[0] is None
     
    # check polygons in m2
    polys2 = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]

    assert 4 <= len(polys2) <= 5

    assert_verts_is_box(polys2[0], [[1, 4], [0, 1]])
    assert_verts_is_box(polys2[1], [[1, 4], [1, 2]])
    assert_verts_is_box(polys2[2], [[1, 4], [2, 3]])
    assert_verts_is_box(polys2[3], [[1, 4], [3, 4]])

def test_agg_with_reset():
    'test the aggregation of states with a reset'

    # m1 dynamics: x' == 1, y' == 0, x0: [-3, -2], y0: [0, 1], step: 1.0
    # m1 invariant: x + y <= 0
    # m1 -> m2 guard: x + y >= 0 and y <= 0.5, reset = [[0, -1, 0], [1, 0, 0]] (x' = -y, y' = x, remove a)
    # m2 dynamics: x' == 0, y' == 0
    # time bound: 4
    # expected result: last state is line (not box!) from (0, 0) to (-0.5, -0.5) 

    ha = HybridAutomaton()

    # mode one: x' = 1, y' = 0, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 0], [0, 0, 0]])

    # mode two: x' = 0, y' = 1 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0], [0, 0]])

    # invariant: x + y <= 0
    m1.set_invariant([[1, 1, 0]], [0])

    # guard: x + y == 0 & y <= 0.5
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard([[-1, -1, 0], [1, 1, 0], [0, 1, 0]], [0, 0, 0.5])
    #trans1.set_reset(np.identity(3)[:2])
    trans1.set_reset(np.array([[0, -1, 0], [1, 0, 0]], dtype=float))

    # initial set has x0 = [-3, -2], y = [0, 1], a = 1
    init_lpi = lputil.from_box([(-3, -2), (0, 1), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 4.0)
    settings.stdout = HylaaSettings.STDOUT_NONE
    
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    # use agg_box
    settings.aggstrat.agg_type = Aggregated.AGG_BOX

    core = Core(ha, settings)
    
    result = core.run(init_list)

    lpi = result.last_cur_state.lpi

    # 2 basis matrix rows, 4 init constraints rows, 6 rows from guard conditions (2 from each)
    assert lpi.get_num_rows() == 2 + 4 + 6

    verts = result.last_cur_state.verts(core.plotman)
    assert len(verts) == 3
    assert np.allclose(verts[0], verts[-1])
    
    assert pair_almost_in((0, 0), verts)
    assert pair_almost_in((-0.5, -0.5), verts)

def test_agg_to_more_vars():
    'test the aggregation of states with a reset to a mode with new variables'

    ha = HybridAutomaton()

    # mode one: x' = 1, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 1], [0, 0]])

    # mode two: x' = 0, a' = 0, y' == 1 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, 0], [0, 0, 0], [0, 1, 0]])

    # invariant: x <= 3.0
    m1.set_invariant([[1, 0]], [3.0])

    # guard: True
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard_true()

    reset_mat = [[1, 0], [0, 1], [0, 0]]
    reset_minkowski = [[0], [0], [1]]
    reset_minkowski_constraints = [[1], [-1]]
    reset_minkowski_rhs = [3, -3] # y0 == 3
    
    trans1.set_reset(reset_mat, reset_minkowski, reset_minkowski_constraints, reset_minkowski_rhs)

    # initial set has x0 = [0, 1], a = 1
    init_lpi = lputil.from_box([(0, 1), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 4.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.store_plot_result = True
    settings.plot.xdim_dir = 0
    settings.plot.ydim_dir = {'m1': 1, 'm2': 2}

    result = Core(ha, settings).run(init_list)

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m1']]

    # 4 steps because invariant is allowed to be false for the final step
    assert 4 <= len(polys) <= 5, "expected invariant to become false after 4/5 steps"

    assert_verts_is_box(polys[0], [[0, 1], [1, 1]])
    assert_verts_is_box(polys[1], [[1, 2], [1, 1]])
    assert_verts_is_box(polys[2], [[2, 3], [1, 1]])
    assert_verts_is_box(polys[3], [[3, 4], [1, 1]])

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]

    assert_verts_is_box(polys[0], [[1, 4], [3, 3]])
    assert_verts_is_box(polys[1], [[1, 4], [4, 4]])

def test_agg_ha():
    'test aggregation with the harmonic oscillator dynamics'

    ha = HybridAutomaton('Deaggregation Example')

    m1 = ha.new_mode('green')
    m1.set_dynamics([[0, 1], [-1, 0]])

    m2 = ha.new_mode('cyan')
    m2.set_dynamics([[0, 0, 0], [0, 0, -2], [0, 0, 0]])

    t1 = ha.new_transition(m1, m2)
    t1.set_guard_true()
    reset_mat = [[1, 0], [0, 1], [0, 0]]
    t1.set_reset(reset_mat, [[0], [0], [1]], [[1], [-1]], [1, -1]) # create 3rd variable with a0 = 1

    mode = ha.modes['green']
    init_lpi = lputil.from_box([(-5, -4), (-0.5, 0.5)], mode)
    
    init_list = [StateSet(init_lpi, mode)]

    step = math.pi/4
    settings = HylaaSettings(step, 2*step)
    settings.process_urgent_guards = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.stdout = HylaaSettings.STDOUT_DEBUG

    core = Core(ha, settings)
    core.setup(init_list)

    core.do_step() # pop
    #xs, ys = zip(*core.cur_state.verts(core.plotman))
    #plt.plot(xs, ys, 'k-')
    
    core.do_step() # 0
    #xs, ys = zip(*core.cur_state.verts(core.plotman))
    #plt.plot(xs, ys, 'k-')

    core.do_step() # 1
    #xs, ys = zip(*core.cur_state.verts(core.plotman))
    #plt.plot(xs, ys, 'k-')
    
    core.do_step() # 2
    assert len(core.aggdag.waiting_list) > 1

    #for state in core.waiting_list:
    #    xs, ys = zip(*state.verts(core.plotman))
    #    plt.plot(xs, ys, 'k-')
    
    core.do_step() # pop
    assert not core.aggdag.waiting_list

    lpi = core.aggdag.get_cur_state().lpi

    # 3 constraints from basis matrix
    # 2 aggregation directions from premode arnoldi, +1 from null space
    # + 2 more aggregation directions from box (3rd is omited since it's exactly the same as null space direction)

    #print(lpi)
    #xs, ys = zip(*core.cur_state.verts(core.plotman))
    #plt.plot(xs, ys, 'r--')
    #plt.show()

    assert lpi.get_num_rows() == 3 + 2 * (5)
    assert lputil.is_point_in_lpi((-5, 2, 1), lpi)

def test_agg_no_counterexample():
    'test that aggregation to error does not create a counterexample'

    # m1 dynamics: x' == 1, y' == 0, x0, y0: [0, 1], step: 1.0
    # m1 invariant: x <= 3
    # m1 -> m2 guard: True
    # m2 dynamics: x' == 0, y' == 1
    # m2 -> error: y >= 3

    ha = HybridAutomaton()

    # mode one: x' = 1, y' = 0, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 0], [0, 0, 0]])

    # mode two: x' = 0, y' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, 0], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 3.0
    m1.set_invariant([[1, 0, 0]], [3.0])

    # guard: True
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard_true()

    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, 'trans2')
    trans2.set_guard([[0, -1, 0]], [-3]) # y >= 3

    # initial set has x0 = [0, 1], t = [0, 1], a = 1
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    result = Core(ha, settings).run(init_list)

    assert not result.counterexample

def test_chull_lines():
    'tests aggregation of two lines in 2d using convex hull'

    mode = HybridAutomaton().new_mode('mode_name')

    center = [-5, -1, 7]
    generator = [0.5, 0.1, 1.0]
    lpi = lputil.from_zonotope(center, [generator], mode)

    t1 = math.pi / 3
    a_mat = np.array([[-0.3, 1, 0], [-1, -0.3, 0], [0, 0.1, 1.1]], dtype=float)
    bm = expm(a_mat * t1)
    lputil.set_basis_matrix(lpi, bm)

    lpi_list = [lpi.clone()]

    all_verts = []
    verts = lpplot.get_verts(lpi)
    all_verts += verts
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k-')

    t2 = t1 + 0.1
    bm = expm(a_mat * t2)
    lputil.set_basis_matrix(lpi, bm)

    lpi_list.append(lpi.clone())

    verts = lpplot.get_verts(lpi)
    all_verts += verts
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k-')

    chull_lpi = lputil.aggregate_chull(lpi_list, mode)

    #xs, ys = zip(*lpplot.get_verts(chull_lpi))
    #plt.plot(xs, ys, 'r--')

    #plt.show()

    for vert in all_verts:
        assert lputil.is_point_in_lpi(vert, chull_lpi)

def test_chull_drivetrain():
    'convex hull aggregation debugging from drivetrain system'

    mode = HybridAutomaton().new_mode('mode_name')

    center = [-0.0432, -11, 0, 30, 0, 30, 360, -0.0013, 30, -0.0013, 30, 0, 1]
    generator = [0.0056, 4.67, 0, 10, 0, 10, 120, 0.0006, 10, 0.0006, 10, 0, 0]

    lpi = lputil.from_zonotope(center, [generator], mode)

    # neg_angle init dynamics
    a_mat = np.array([ \
        [0, 0, 0, 0, 0, 0, 0.0833333333333333, 0, -1, 0, 0, 0, 0], \
        [13828.8888888889, -26.6666666666667, 60, 60, 0, 0, -5, -60, 0, 0, 0, 0, 116.666666666667], \
        [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -5], \
        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, -714.285714285714, -0.04, 0, 0, 0, 714.285714285714, 0, 0, 0], \
        [-2777.77777777778, 3.33333333333333, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -83.3333333333333], \
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0], \
        [100, 0, 0, 0, 0, 0, 0, -1000, -0.01, 1000, 0, 0, 3], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0], \
        [0, 0, 0, 0, 1000, 0, 0, 1000, 0, -2000, -0.01, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        ], dtype=float)

    # plot dimensions
    xdim = 0
    ydim = 1

    step = 5.0E-2
    t1 = 0
    bm = expm(a_mat * t1)
    lputil.set_basis_matrix(lpi, bm)

    lpi_list = [lpi.clone()]

    all_verts = []
    verts = lpplot.get_verts(lpi, xdim=xdim, ydim=ydim)
    all_verts += verts
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k-')

    t2 = t1 + step
    bm = expm(a_mat * t2)
    lputil.set_basis_matrix(lpi, bm)

    lpi_list.append(lpi.clone())

    verts = lpplot.get_verts(lpi, xdim=xdim, ydim=ydim)
    all_verts += verts
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k-')

    chull_lpi = lputil.aggregate_chull(lpi_list, mode)

    plot_vecs = lpplot.make_plot_vecs(num_angles=256, offset=0.01)
    verts = lpplot.get_verts(chull_lpi, xdim=xdim, ydim=ydim, plot_vecs=plot_vecs)
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'r--')

    #plt.show()

    for vert in all_verts:
        assert lputil.is_point_in_lpi(vert, chull_lpi)

def test_chull_one_step_inputs():
    'test convex hull with one-step lpi for a system with inputs (bug where current vars was not set correctly)'

    mode = HybridAutomaton().new_mode('mode_name')

    step_size = math.pi/4

    a_mat = np.array([[0, 1], [-1, 0]], dtype=float)

    b_mat = [[1], [0]]
    b_constraints = [[1], [-1]]
    b_rhs = [0.2, 0.2]

    mode.set_dynamics(a_mat)
    mode.set_inputs(b_mat, b_constraints, b_rhs)
    mode.init_time_elapse(step_size)

    box = [[-5, -4], [0.0, 1.0]]
    lpi = lputil.from_box(box, mode)

    lpi_one_step = lpi.clone()
    bm, ie_mat = mode.time_elapse.get_basis_matrix(1)

    lputil.set_basis_matrix(lpi_one_step, bm)
    lputil.add_input_effects_matrix(lpi_one_step, ie_mat, mode)

    lpi_list = [lpi, lpi_one_step]
    chull_lpi = lputil.aggregate_chull(lpi_list, mode)

    # 2 current vars and 2 total input effect vars, so expected to be 4 from the end
    assert chull_lpi.cur_vars_offset == chull_lpi.get_num_cols() - 4, "cur_vars in wrong place"



'''
Tests for LP operations. Made for use with py.test
'''

import math

import numpy as np

from scipy.linalg import expm
from scipy.sparse import csr_matrix, csc_matrix
from scipy.io import loadmat

import swiglpk as glpk

from hylaa import lputil, lpplot
from hylaa.hybrid_automaton import HybridAutomaton, LinearConstraint

from util import assert_verts_is_box, pair_almost_in

import matplotlib.pyplot as plt

def test_from_box():
    'tests from_box'
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    assert lpi.basis_mat_pos == (0, 0)
    assert lpi.dims == 2

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0], \
        [0, 1, 0, -1], \
        [-1, 0, 0, 0], \
        [1, 0, 0, 0], \
        [0, -1, 0, 0], \
        [0, 1, 0, 0]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

def test_print_lp():
    'test printing the lp to stdout'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))
    assert str(lpi) is not None

def test_set_basis_matrix():
    'tests lputil set_basis_matrix on harmonic oscillator example'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    basis = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    assert np.allclose(lputil.get_basis_matrix(lpi), basis)

    mat, vec = lpi.get_full_constraints(), lpi.get_rhs()

    expected_mat = np.array([\
        [0, 1, -1, 0], \
        [-1, 0, 0, -1], \
        [-1, 0, 0, 0], \
        [1, 0, 0, 0], \
        [0, -1, 0, 0], \
        [0, 1, 0, 0]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1], dtype=float)

    assert np.allclose(vec, expected_vec)

    assert np.allclose(mat.toarray(), expected_mat)

def test_check_intersection():
    'tests check_intersection on the harmonic oscillator example'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    # check if initially y >= 4.5 is possible (should be false)
    direction = np.array([0, -1], dtype=float)
    lc = LinearConstraint(direction, -4.5)

    assert not lputil.check_intersection(lpi, lc)

    # after basis matrix update
    basis = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    # now check if y >= 4.5 is possible (should be true)
    assert lputil.check_intersection(lpi, lc)

def test_verts():
    'tests verts'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))
    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)

    assert_verts_is_box(verts, [(-5, -4), (0, 1)])

def test_add_init_constraint():
    'tests add_init_constraint on the harmonic oscillator example'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    # update basis matrix
    basis_mat = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis_mat)

    # minimize y should give 4.0
    miny = lpi.minimize([0, 1], columns=[lpi.cur_vars_offset + 1])[0]
    assert abs(miny - 4.0) < 1e-6

    # add constraint: y >= 4.5
    direction = np.array([0, -1], dtype=float)

    new_row = lputil.add_init_constraint(lpi, direction, -4.5)

    assert new_row == 6, "new constraint should have been added in row index 6"

    # minimize y should give 4.5
    miny = lpi.minimize([0, 1], columns=[lpi.cur_vars_offset + 1])[0]
    assert abs(miny - 4.5) < 1e-6

    # check verts()
    verts = lpplot.get_verts(lpi)

    assert len(verts) == 5
    
    assert [0.0, 5.0] in verts
    assert [1.0, 5.0] in verts
    assert [0.0, 4.5] in verts
    assert [1.0, 4.5] in verts
    assert verts[0] == verts[-1]

def test_replace_init_constraint():
    'tests try_replace_init_constraint on the harmonic oscillator example'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    # update basis matrix
    basis_mat = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis_mat)

    # minimize y should give 4.0
    miny = lpi.minimize([0, 1], columns=[lpi.cur_vars_offset + 1])[0]
    assert abs(miny - 4.0) < 1e-6

    # add constraint: y >= 4.5
    direction = np.array([0, -1], dtype=float)

    row_index = lputil.add_init_constraint(lpi, direction, -4.5)
    assert lpi.get_rhs()[-1] == -4.5

    # minimize y should give 4.5
    miny = lpi.minimize([0, 1], columns=[lpi.cur_vars_offset + 1])[0]
    assert abs(miny - 4.5) < 1e-6

    assert lpi.get_num_rows() == 7

    # try to replace constraint y >= 4.6 (should be stronger than 4.5)
    row_index, is_stronger = lputil.try_replace_init_constraint(lpi, row_index, direction, -4.6)

    assert is_stronger
    assert row_index == 6
    assert lpi.get_num_rows() == 7
    assert lpi.get_rhs()[row_index] == -4.6

    # try to replace constraint x <= 0.9 (should be incomparable)
    xdir = np.array([1, 0], dtype=float)
    row_index, is_stronger = lputil.try_replace_init_constraint(lpi, row_index, xdir, 0.9)

    assert not is_stronger
    assert lpi.get_num_rows() == 8
    assert lpi.get_rhs()[row_index] == 0.9

    # check verts()
    verts = lpplot.get_verts(lpi)

    assert len(verts) == 5
    
    assert [0.0, 5.0] in verts
    assert [0.9, 5.0] in verts
    assert [0.0, 4.6] in verts
    assert [0.9, 4.6] in verts
    assert verts[0] == verts[-1]

def test_get_basis_matrix():
    'tests lputil get_basis_matrix on harmonic oscillator example'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    basis = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    mat = lputil.get_basis_matrix(lpi)

    assert np.allclose(mat, basis)

def test_add_curtime_constraints():
    'tests add_curtime_constraints'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    # new constraint to be added, x <= 3.14, y <= 10
    csr_constraint = csr_matrix(np.array([[1, 0], [0, 1]], dtype=float))
    rhs = np.array([3.14, 10], dtype=float)

    lputil.add_curtime_constraints(lpi, csr_constraint, rhs)

    mat = lpi.get_full_constraints()
    vec = lpi.get_rhs()
    types = lpi.get_types()

    expected_mat = np.array([\
        [1, 0, -1, 0], \
        [0, 1, 0, -1], \
        [-1, 0, 0, 0], \
        [1, 0, 0, 0], \
        [0, -1, 0, 0], \
        [0, 1, 0, 0], \
        [0, 0, 1, 0], \
        [0, 0, 0, 1]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1, 3.14, 10], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, up, up]

    assert np.allclose(vec, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)

def test_add_reset_variables():
    'tests add_reset_variables'
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    reset_csr = csr_matrix(2 * np.identity(2))
    mode_id = 1
    transition_id = 13
    lputil.add_reset_variables(lpi, mode_id, transition_id, reset_csr=reset_csr)

    assert lpi.dims == 2

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0, 0, 0, 0, 0], \
        [0, 1, 0, -1, 0, 0, 0, 0], \
        [-1, 0, 0, 0, 0, 0, 0, 0], \
        [1, 0, 0, 0, 0, 0, 0, 0], \
        [0, -1, 0, 0, 0, 0, 0, 0], \
        [0, 1, 0, 0, 0, 0, 0, 0], \
        [0, 0, 2, 0, -1, 0, 0, 0], \
        [0, 0, 0, 2, 0, -1, 0, 0], \
        [0, 0, 0, 0, 1, 0, -1, 0], \
        [0, 0, 0, 0, 0, 1, 0, -1]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1, 0, 0, 0, 0], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, fx, fx, fx, fx]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "m1_i0_t13", "m1_i1", "m1_c0", "m1_c1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    assert lpi.basis_mat_pos == (8, 4)

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))
    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)

    assert len(verts) == 5
    
    assert [-10.0, 0.] in verts
    assert [-10.0, 2.] in verts
    assert [-8.0, 2.] in verts
    assert [-8.0, 0.] in verts
    assert verts[0] == verts[-1]

    # update the basis matrix to rotate quarter circle
    basis = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)
    assert len(verts) == 5
    
    assert [0.0, 10.] in verts
    assert [0.0, 8.] in verts
    assert [2.0, 8.] in verts
    assert [2.0, 10.] in verts
    assert verts[0] == verts[-1]

def test_add_reset_inputs():
    'tests add_reset_variables'

    mode = HybridAutomaton().new_mode('mode_name')
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], mode)

    reset_csr = csr_matrix(2 * np.identity(2))
    mode_id = 1
    transition_id = 13
    lputil.add_reset_variables(lpi, mode_id, transition_id, reset_csr=reset_csr, successor_has_inputs=True)

    assert lpi.dims == 2

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0, 0, 0, 0, 0, 0, 0], \
        [0, 1, 0, -1, 0, 0, 0, 0, 0, 0], \
        [-1, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, -1, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 2, 0, -1, 0, 0, 0, 0, 0], \
        [0, 0, 0, 2, 0, -1, 0, 0, 0, 0], \
        [0, 0, 0, 0, 1, 0, -1, 0, 1, 0], \
        [0, 0, 0, 0, 0, 1, 0, -1, 0, 1], \
        [0, 0, 0, 0, 0, 0, 0, 0, -1, 0], \
        [0, 0, 0, 0, 0, 0, 0, 0, 0, -1]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1, 0, 0, 0, 0, 0, 0], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, fx, fx, fx, fx, fx, fx]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "m1_i0_t13", "m1_i1", "m1_c0", "m1_c1", "m1_ti0", "m1_ti1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    assert lpi.basis_mat_pos == (8, 4)

    assert lpi.input_effects_offsets == (10, 8)

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))
    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)

    assert len(verts) == 5
    
    assert [-10.0, 0.] in verts
    assert [-10.0, 2.] in verts
    assert [-8.0, 2.] in verts
    assert [-8.0, 0.] in verts
    assert verts[0] == verts[-1]

def test_reset_less_dims():
    '''tests a reset to a mode with less dimensions
    project onto just the y variable multiplied by 0.5
    '''
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))
    assert lpi.dims == 2

    reset_csr = csr_matrix(np.array([[0, 0.5]], dtype=float))
    mode_id = 1
    transition_id = 13
    lputil.add_reset_variables(lpi, mode_id, transition_id, reset_csr=reset_csr)

    assert lpi.dims == 1

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0, 0, 0], \
        [0, 1, 0, -1, 0, 0], \
        [-1, 0, 0, 0, 0, 0], \
        [1, 0, 0, 0, 0, 0], \
        [0, -1, 0, 0, 0, 0], \
        [0, 1, 0, 0, 0, 0], \
        [0, 0, 0, 0.5, -1, 0], \
        [0, 0, 0, 0, 1, -1]], dtype=float)

    expected_vec = np.array([0, 0, 5, -4, 0, 1, 0, 0], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, fx, fx]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "m1_i0_t13", "m1_c0"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    assert lpi.basis_mat_pos == (7, 4)
    assert lpi.dims == 1

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))

    verts = lpplot.get_verts(lpi, xdim=0, ydim=None, plot_vecs=plot_vecs, cur_time=0)

    assert len(verts) == 3
    
    assert [0.5, 0] in verts
    assert [0, 0] in verts
    assert verts[0] == verts[-1]

    # update the basis matrix
    basis = np.array([[2]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    verts = lpplot.get_verts(lpi, xdim=0, ydim=None, plot_vecs=plot_vecs, cur_time=0)
    assert len(verts) == 3

    assert [1.0, 0] in verts
    assert [0, 0] in verts
    assert verts[0] == verts[-1]

def test_reset_minkowski():
    '''tests reset with a minkowski sum term and a new variable

    pre reset we have x = [-5, -4], y = [0, 1]
    post reset we have x = [-15, -14] (-10), y = [0, 1], t' = [0, 5]

    reset_matrix is [[1, 0], [0, 1], [0, 0]]
    minkowski_csr is [[1, 0], [0, 0], [0, 1]]
    minkowski_constraints_csr is [[1, 0], [-1, 0], [0, 1], [0, -1]]
    minkowski_constraints_rhs is [-10, 10, 5, 0]

    '''
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    reset_csr = csr_matrix([[1, 0], [0, 1], [0, 0]], dtype=float)
    mode_id = 1
    transition_id = 13

    minkowski_csr = csr_matrix([[1, 0], [0, 0], [0, 1]], dtype=float)
    constraints_csr = csr_matrix([[1, 0], [-1, 0], [0, 1], [0, -1]], dtype=float)
    constraints_rhs = np.array([-10, 10, 5, 0], dtype=float)
    
    lputil.add_reset_variables(lpi, mode_id, transition_id, reset_csr=reset_csr, minkowski_csr=minkowski_csr, \
                               minkowski_constraints_csr=constraints_csr, minkowski_constraints_rhs=constraints_rhs)

    assert lpi.dims == 3

    # basis matrix should be at 9, 6
    assert lpi.basis_mat_pos == (9, 6)

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "reset0", "reset1", "m1_i0_t13", "m1_i1", "m1_i2", \
                          "m1_c0", "m1_c1", "m1_c2"]

    assert lpi.get_names() == expected_names

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))
    verts = lpplot.get_verts(lpi, xdim=0, ydim=1, plot_vecs=plot_vecs)

    assert len(verts) == 5
    
    assert [-15.0, 0.] in verts
    assert [-15.0, 1.] in verts
    assert [-14.0, 1.] in verts
    assert [-14.0, 0.] in verts
    assert verts[0] == verts[-1]

    verts = lpplot.get_verts(lpi, xdim=2, ydim=None, plot_vecs=plot_vecs, cur_time=0.0)

    assert len(verts) == 3
    
    assert [0, 0.] in verts
    assert [5, 0.] in verts
    assert verts[0] == verts[-1]

    lputil.set_basis_matrix(lpi, 3 * np.identity(3))

    verts = lpplot.get_verts(lpi, xdim=2, ydim=None, plot_vecs=plot_vecs, cur_time=0.0)

    assert len(verts) == 3
    
    assert [0, 0.] in verts
    assert [15, 0.] in verts
    assert verts[0] == verts[-1]

def test_init_triangle():
    'tests initialization from a non-box initial set of states'

    # x + y < 1, x > 0, y > 0

    constraints_mat = [[1, 1], [-1, 0], [0, -1]]
    constraints_rhs = [1, 0, 0]

    lpi = lputil.from_constraints(constraints_mat, constraints_rhs, HybridAutomaton().new_mode('mode_name'))

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0], \
        [0, 1, 0, -1], \
        [1, 1, 0, 0], \
        [-1, 0, 0, 0], \
        [0, -1, 0, 0]], dtype=float)

    expected_vec = np.array([0, 0, 1, 0, 0], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    # check verts

    plot_vecs = lpplot.make_plot_vecs(4, offset=(math.pi / 4.0))
    verts = lpplot.get_verts(lpi, plot_vecs=plot_vecs)

    assert len(verts) == 4
    
    assert [0., 1.] in verts
    assert [0., 0.] in verts
    assert [1., 0] in verts
    assert verts[0] == verts[-1]

def test_get_box_center():
    'test get_box_center'

    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    pt = lputil.get_box_center(lpi)
    assert len(pt) == 2
    assert abs(pt[0] - (-4.5)) < 1e-4
    assert abs(pt[1] - (0.5)) < 1e-4

    basis = np.array([[0, 1], [-1, 0]], dtype=float)
    lputil.set_basis_matrix(lpi, basis)

    pt = lputil.get_box_center(lpi)
    assert len(pt) == 2
    assert abs(pt[0] - (0.5)) < 1e-4
    assert abs(pt[1] - (4.5)) < 1e-4

    # try it rotated 1/4 around the circle
    a_mat = np.array([[0, 1], [-1, 0]], dtype=float)

    bm = expm(a_mat * math.pi / 4)
    lputil.set_basis_matrix(lpi, bm)

    expected = np.dot(bm, np.array([[-4.5], [0.5]], dtype=float))

    pt = lputil.get_box_center(lpi)

    assert len(pt) == 2
    assert abs(pt[0] - expected[0][0]) < 1e-4
    assert abs(pt[1] - expected[1][0]) < 1e-4


def test_make_direction_matrix():
    '''
    test making the direction matrix on a 2d example. The first vector should match the dynamics as the passed-in 
    point, while the second should be orthogonal to the first
    '''

    a_csr = csr_matrix(np.array([[1, 0], [0, 1]], dtype=float))
    pt = [2, 2] # derivative should be <2, 2>

    mat = lputil.make_direction_matrix(pt, a_csr)

    assert mat.shape == (2, 2)

    # first row should be <sqrt(2)/2, sqrt(2)/2> (normalized version of derivative)
    assert abs(mat[0][0] - math.sqrt(2)/2) < 1e-6
    assert abs(mat[0][1] - math.sqrt(2)/2) < 1e-6

    for a, row_a in enumerate(mat):
        assert abs(np.linalg.norm(row_a) - 1.0) < 1e-6, "rows should be normalized"

        for b, row_b in enumerate(mat):
            if a == b:
                continue

            assert np.dot(row_a, row_b) < 1e-6, "rows should be orthononal"
    
def test_direction_matrix_empty():
    '''test make_direction matrix when the dynamics matrix is empty

    it should just return any orthonormal set of vectors of the appropriate dimension
    '''

    dims = 5
    a_csr = csr_matrix(np.zeros((dims, dims)))
    pt = [0] * dims

    mat = lputil.make_direction_matrix(pt, a_csr)

    assert mat.shape == (dims, dims)

    for a, row_a in enumerate(mat):
        assert abs(np.linalg.norm(row_a) - 1.0) < 1e-6, "rows should be normalized"

        for b, row_b in enumerate(mat):
            if a == b:
                continue

            assert np.dot(row_a, row_b) < 1e-6, "rows should be orthononal"



def test_reject_constant_inputs():
    'tests the detection of B matrix + constraints where an input is fixed to a constant'

    # x' = Ax + Bu
    # x: [[1, 0], [0, 1]]

    mode = HybridAutomaton().new_mode('mode_name')
    mode.set_dynamics(np.identity(2))

    b_mat = np.identity(2)
    b_con = [[1, 0], [-1, 0], [0, -1], [0, -1]]
    b_rhs = [1, 0, 2, -2]

    try:
        mode.set_inputs(b_mat, b_con, b_rhs)
        assert False, "expected fixed inputs to be rejected"
    except AssertionError:
        pass

    b_rhs = [1, 0, 2, -1]
    mode.set_inputs(b_mat, b_con, b_rhs)
    # should be okay

    b_mat = [[1, 1], [1, 1]]
    b_rhs = [1, 0, 2, -2]
    mode.set_inputs(b_mat, b_con, b_rhs)
    # should be okay (b_mat is not identity)

    b_mat = np.identity(2)
    b_con = [[1, 1], [-1, -1], [1, -1], [-1, 1]]
    b_rhs = [1, 0, 2, -2]
    mode.set_inputs(b_mat, b_con, b_rhs)
    # should be okay

    b_mat = np.identity(2)
    b_con = [[1, 1], [-1, -1], [1, -1], [-1, 1]]
    b_rhs = [1, -2, 2, -2]
    try:
        mode.set_inputs(b_mat, b_con, b_rhs)
        assert False, "expected unsat inputs to be rejected"
    except AssertionError:
        pass

def test_box_inputs():
    'tests from_box with a simple input effects matrix'

    # x' = Ax + Bu
    # A = 0
    # B = [[1, 0], [0, 2]]
    # u1 and u2 are bounded between [1, 10]

    # (init) step 0: [0, 1] x [0, 1]
    # step 1: [1, 11] x [2, 21]
    # step 2: [2, 21] x [4, 41]

    mode = HybridAutomaton().new_mode('mode_name')
    mode.set_dynamics(np.zeros((2, 2)))
    mode.set_inputs([[1, 0], [0, 2]], [[1, 0], [-1, 0], [0, 1], [0, -1]], [10, -1, 10, -1])

    init_box = [[0, 1], [0, 1]]
    lpi = lputil.from_box(init_box, mode)

    assert lpi.basis_mat_pos == (0, 0)
    assert lpi.dims == 2
    assert lpi.cur_vars_offset == 2
    assert lpi.input_effects_offsets == (6, 4) # row 6, column 4 for total input effects offsets

    # step 0
    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0, 1, 0], \
        [0, 1, 0, -1, 0, 1], \
        [-1, 0, 0, 0, 0, 0], \
        [1, 0, 0, 0, 0, 0], \
        [0, -1, 0, 0, 0, 0], \
        [0, 1, 0, 0, 0, 0], \
        [0, 0, 0, 0, -1, 0], \
        [0, 0, 0, 0, 0, -1]], dtype=float)

    expected_vec = np.array([0, 0, 0, 1, 0, 1, 0, 0], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, fx, fx]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "m0_ti0", "m0_ti1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    verts = lpplot.get_verts(lpi)
    assert_verts_is_box(verts, init_box)

    # do step 1
    mode.init_time_elapse(1.0)
    basis_mat, input_mat = mode.time_elapse.get_basis_matrix(1)

    lputil.set_basis_matrix(lpi, basis_mat)
    lputil.add_input_effects_matrix(lpi, input_mat, mode)

    mat = lpi.get_full_constraints()
    types = lpi.get_types()
    rhs = lpi.get_rhs()
    names = lpi.get_names()

    expected_mat = np.array([\
        [1, 0, -1, 0, 1, 0, 0, 0], \
        [0, 1, 0, -1, 0, 1, 0, 0], \
        [-1, 0, 0, 0, 0, 0, 0, 0], \
        [1, 0, 0, 0, 0, 0, 0, 0], \
        [0, -1, 0, 0, 0, 0, 0, 0], \
        [0, 1, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, -1, 0, 1, 0], \
        [0, 0, 0, 0, 0, -1, 0, 2], \
        [0, 0, 0, 0, 0, 0, 1, 0], \
        [0, 0, 0, 0, 0, 0, -1, 0], \
        [0, 0, 0, 0, 0, 0, 0, 1], \
        [0, 0, 0, 0, 0, 0, 0, -1]], dtype=float)

    expected_vec = np.array([0, 0, 0, 1, 0, 1, 0, 0, 10, -1, 10, -1], dtype=float)

    fx = glpk.GLP_FX
    up = glpk.GLP_UP
    expected_types = [fx, fx, up, up, up, up, fx, fx, up, up, up, up]

    expected_names = ["m0_i0", "m0_i1", "m0_c0", "m0_c1", "m0_ti0", "m0_ti1", "m0_I0", "m0_I1"]

    assert np.allclose(rhs, expected_vec)
    assert types == expected_types
    assert np.allclose(mat.toarray(), expected_mat)
    assert names == expected_names

    verts = lpplot.get_verts(lpi)

    assert_verts_is_box(verts, [(1, 11), (2, 21)])

    # do step 2
    basis_mat, input_mat = mode.time_elapse.get_basis_matrix(2)
    lputil.set_basis_matrix(lpi, basis_mat)
    lputil.add_input_effects_matrix(lpi, input_mat, mode)

    verts = lpplot.get_verts(lpi)
    assert_verts_is_box(verts, [(2, 21), (4, 41)])

def test_bloat():
    'tests bloat'
    
    lpi = lputil.from_box([[-5, -4], [0, 1]], HybridAutomaton().new_mode('mode_name'))

    lputil.bloat(lpi, 0.5)

    verts = lpplot.get_verts(lpi)

    assert_verts_is_box(verts, [(-5.5, -3.5), (-0.5, 1.5)])

def test_scale():
    'tests scale'
    
    lpi = lputil.from_box([[4, 5], [-1, 1]], HybridAutomaton().new_mode('mode_name'))

    lputil.scale_with_bm(lpi, 2.0)

    verts = lpplot.get_verts(lpi)

    assert_verts_is_box(verts, [(8, 10), (-2, 2)])

def test_minkowski_sum_box():
    'tests minkowski_sum with 2 box sets'

    mode = HybridAutomaton().new_mode('mode_name')
    
    lpi1 = lputil.from_box([[-1, 1], [-2, 2]], mode)
    lpi2 = lputil.from_box([[-.1, .1], [-.2, .2]], mode)

    lpi = lputil.minkowski_sum([lpi1, lpi2], mode)

    verts = lpplot.get_verts(lpi)

    assert_verts_is_box(verts, [(-1.1, 1.1), (-2.2, 2.2)])

def test_minkowski_box_diamond():
    'tests minkowski_sum of a box and a diamond'

    mode = HybridAutomaton().new_mode('mode_name')
    
    lpi1 = lputil.from_box([[-1, 1], [-1, 1]], mode)

    # -1 <= x + y <= 1
    # -1 <= x - y <= 1
    constraints_mat = [[1, 1], [-1, -1], [1, -1], [-1, 1]]
    constraints_rhs = [1, 1, 1, 1]
    lpi2 = lputil.from_constraints(constraints_mat, constraints_rhs, mode)
    
    #verts = lpplot.get_verts(lpi1)
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'r--')

    #verts = lpplot.get_verts(lpi2)
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'b--')

    lpi = lputil.minkowski_sum([lpi1, lpi2], mode)

    #verts = lpplot.get_verts(lpi)
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k:')
    
    #plt.show()

    verts = lpplot.get_verts(lpi)
    assert len(verts) == 9 # octogon + wrap

    expected = [(-2, 1), (-2, -1), (-1, -2), (1, -2), (2, -1), (2, 1), (1, 2), (-1, 2)]

    for pt in expected:
        assert pair_almost_in(pt, verts), f"{pt} not found in verts: {verts}"

def test_from_input_constraints():
    'test making an lpi set from input constraints'

    mode = HybridAutomaton().new_mode('mode_name')

    b_mat = [[1], [0]]
    b_constraints = [[1], [-1]]
    b_rhs = [0.2, 0.2]

    # result should have two vertices, at (-0.2, 0) and (0.2, 0)
    lpi = lputil.from_input_constraints(b_mat, b_constraints, b_rhs, mode)

    print(lpi)

    assert lpi.get_num_cols() == 5
    assert lpi.cur_vars_offset == 3
    
    verts = lpplot.get_verts(lpi)
    assert len(verts) == 3 # 2 + wrap

    expected = [(-0.2, 0), (0.2, 0)]

    for pt in expected:
        assert pair_almost_in(pt, verts), f"{pt} not found in verts: {verts}"
